# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"""Unit tests for steno_dictionary.py."""

import os
import stat
import tempfile

import pytest

from plover.steno_dictionary import StenoDictionary, StenoDictionaryCollection


def test_dictionary():
    notifications = []
    def listener(longest_key):
        notifications.append(longest_key)

    d = StenoDictionary()
    assert d.longest_key == 0

    d.add_longest_key_listener(listener)
    d[('S',)] = 'a'
    assert d.longest_key == 1
    assert notifications == [1]
    d[('S', 'S', 'S', 'S')] = 'b'
    assert d.longest_key == 4
    assert notifications == [1, 4]
    d[('S', 'S')] = 'c'
    assert d.longest_key == 4
    assert d[('S', 'S')] == 'c'
    assert notifications == [1, 4]
    del d[('S', 'S', 'S', 'S')]
    assert d.longest_key == 2
    assert notifications == [1, 4, 2]
    del d[('S',)]
    assert d.longest_key == 2
    assert notifications == [1, 4, 2]
    assert d.reverse_lookup('c') == [('S', 'S')]
    assert d.casereverse_lookup('c') == ['c']
    d.clear()
    assert d.longest_key == 0
    assert notifications == [1, 4, 2, 0]
    assert d.reverse_lookup('c') == []
    assert d.casereverse_lookup('c') == []

    d.remove_longest_key_listener(listener)
    d[('S', 'S')] = 'c'
    assert d.longest_key == 2
    assert notifications == [1, 4, 2, 0]


def test_dictionary_collection():
    d1 = StenoDictionary()
    d1[('S',)] = 'a'
    d1[('T',)] = 'b'
    d1.path = 'd1'
    d2 = StenoDictionary()
    d2[('S',)] = 'c'
    d2[('W',)] = 'd'
    d2.path = 'd2'
    dc = StenoDictionaryCollection([d2, d1])
    assert dc.lookup(('S',)) == 'c'
    assert dc.lookup(('W',)) == 'd'
    assert dc.lookup(('T',)) == 'b'
    f = lambda k, v: v == 'c'
    dc.add_filter(f)
    assert dc.lookup(('S',)) is None
    assert dc.raw_lookup(('S',)) == 'c'
    assert dc.lookup(('W',)) == 'd'
    assert dc.lookup(('T',)) == 'b'
    assert dc.reverse_lookup('c') == [('S',)]

    dc.remove_filter(f)
    assert dc.lookup(('S',)) == 'c'
    assert dc.lookup(('W',)) == 'd'
    assert dc.lookup(('T',)) == 'b'

    assert dc.reverse_lookup('c') == [('S',)]

    dc.set(('S',), 'e')
    assert dc.lookup(('S',)) == 'e'
    assert d2[('S',)] == 'e'

    dc.set(('S',), 'f', path='d1')
    assert dc.lookup(('S',)) == 'e'
    assert d1[('S',)] == 'f'
    assert d2[('S',)] == 'e'

    # Iterating on a StenoDictionaryCollection is
    # the same as iterating on its dictionaries' paths.
    assert list(dc) == ['d2', 'd1']

    # Test get and [].
    assert dc.get('d1') == d1
    assert dc['d1'] == d1
    assert dc.get('invalid') is None
    with pytest.raises(KeyError):
        dc['invalid']


def test_dictionary_collection_writeable():
    d1 = StenoDictionary()
    d1[('S',)] = 'a'
    d1[('T',)] = 'b'
    d2 = StenoDictionary()
    d2[('S',)] = 'c'
    d2[('W',)] = 'd'
    d2.readonly = True
    dc = StenoDictionaryCollection([d2, d1])
    assert dc.first_writable() == d1
    dc.set(('S',), 'A')
    assert d1[('S',)] == 'A'
    assert d2[('S',)] == 'c'


def test_dictionary_collection_longest_key():

    k1 = ('S',)
    k2 = ('S', 'T')
    k3 = ('S', 'T', 'R')

    dc = StenoDictionaryCollection()
    assert dc.longest_key == 0

    d1 = StenoDictionary()
    d1.path = 'd1'
    d1[k1] = 'a'

    dc.set_dicts([d1])
    assert dc.longest_key == 1

    d1[k2] = 'a'
    assert dc.longest_key == 2

    d2 = StenoDictionary()
    d2.path = 'd2'
    d2[k3] = 'c'

    dc.set_dicts([d2, d1])
    assert dc.longest_key == 3

    del d1[k2]
    assert dc.longest_key == 3

    dc.set_dicts([d1])
    assert dc.longest_key == 1

    dc.set_dicts([])
    assert dc.longest_key == 0


def test_casereverse_del():
    d = StenoDictionary()
    d[('S-G',)] = 'something'
    d[('SPH-G',)] = 'something'
    assert d.casereverse_lookup('something') == ['something']
    del d[('S-G',)]
    assert d.casereverse_lookup('something') == ['something']
    del d[('SPH-G',)]
    assert d.casereverse_lookup('something') == []


def test_reverse_lookup():
    dc = StenoDictionaryCollection()

    d1 = StenoDictionary()
    d1[('PWAOUFL',)] = 'beautiful'
    d1[('WAOUFL',)] = 'beautiful'

    d2 = StenoDictionary()
    d2[('PW-FL',)] = 'beautiful'

    d3 = StenoDictionary()
    d3[('WAOUFL',)] = 'not beautiful'

    # Simple test.
    dc.set_dicts([d1])
    assert dc.reverse_lookup('beautiful') == [('PWAOUFL',), ('WAOUFL',)]

    # No duplicates.
    d2_copy = StenoDictionary()
    d2_copy.update(d2)
    dc.set_dicts([d2_copy, d2])
    assert dc.reverse_lookup('beautiful') == [('PW-FL',)]

    # Don't stop at the first dictionary with matches.
    dc.set_dicts([d2, d1])
    assert dc.reverse_lookup('beautiful') == [('PW-FL',), ('PWAOUFL',), ('WAOUFL',)]

    # Ignore keys overridden by a higher precedence dictionary.
    dc.set_dicts([d3, d2, d1])
    assert dc.reverse_lookup('beautiful') == [('PW-FL',), ('PWAOUFL',)]


def test_dictionary_enabled():
    dc = StenoDictionaryCollection()
    d1 = StenoDictionary()
    d1.path = 'd1'
    d1[('TEFT',)] = 'test1'
    d1[('TEFGT',)] = 'Testing'
    d2 = StenoDictionary()
    d2[('TEFT',)] = 'test2'
    d2[('TEFT', '-G')] = 'Testing'
    d2.path = 'd2'
    dc.set_dicts([d2, d1])
    assert dc.lookup(('TEFT',)) == 'test2'
    assert dc.raw_lookup(('TEFT',)) == 'test2'
    assert dc.casereverse_lookup('testing') == ['Testing']
    assert dc.reverse_lookup('Testing') == [('TEFT', '-G'), ('TEFGT',)]
    d2.enabled = False
    assert dc.lookup(('TEFT',)) == 'test1'
    assert dc.raw_lookup(('TEFT',)) == 'test1'
    assert dc.casereverse_lookup('testing') == ['Testing']
    assert dc.reverse_lookup('Testing') == [('TEFGT',)]
    d1.enabled = False
    assert dc.lookup(('TEST',)) is None
    assert dc.raw_lookup(('TEFT',)) is None
    assert dc.casereverse_lookup('testing') is None
    assert dc.reverse_lookup('Testing') == []


def test_dictionary_readonly():
    class FakeDictionary(StenoDictionary):
        def _load(self, filename):
            pass
        def _save(self):
            raise NotImplementedError
    tf = tempfile.NamedTemporaryFile(delete=False)
    try:
        tf.close()
        d = FakeDictionary.load(tf.name)
        # Writable file: not readonly.
        assert not d.readonly
        # Readonly file: readonly dictionary.
        os.chmod(tf.name, stat.S_IREAD)
        d = FakeDictionary.load(tf.name)
        assert d.readonly
    finally:
        # Deleting the file will fail on Windows
        # if we don't restore write permission.
        os.chmod(tf.name, stat.S_IWRITE)
        os.unlink(tf.name)
    # Assets are always readonly.
    d = FakeDictionary.load('asset:plover:assets/main.json')
    assert d.readonly



import ast
import functools
import operator
import re
import textwrap

from plover import system
from plover.formatting import Formatter
from plover.steno import Stroke, normalize_steno
from plover.steno_dictionary import StenoDictionary
from plover.translation import Translator


class CaptureOutput:

    def __init__(self):
        self.instructions = []
        self.text = ''

    def send_backspaces(self, n):
        assert n <= len(self.text)
        self.text = self.text[:-n]
        self.instructions.append(('b', n))

    def send_string(self, s):
        self.text += s
        self.instructions.append(('s', s))

    def send_key_combination(self, c):
        self.instructions.append(('c', c))

    def send_engine_command(self, c):
        self.instructions.append(('e', c))


def steno_to_stroke(steno):
    if steno_to_stroke.system != system.NAME:
        keys = []
        letters = ''
        has_hyphen = False
        for k in system.KEYS:
            if not has_hyphen and k.startswith('-'):
                has_hyphen = True
                keys.append(None)
                letters += '-'
            keys.append(k)
            letters += k.strip('-')
        steno_to_stroke.keys = keys
        steno_to_stroke.letters = letters
        steno_to_stroke.system = system.NAME
        steno_to_stroke.numbers = {
            v.strip('-'): k.strip('-')
            for k, v in system.NUMBERS.items()
        }
    n = -1
    keys = set()
    for l in steno:
        rl = steno_to_stroke.numbers.get(l)
        if rl is not None:
            keys.add('#')
            l = rl
        n = steno_to_stroke.letters.find(l, n + 1)
        assert n >= 0, (steno_to_stroke.letters, l, n)
        k = steno_to_stroke.keys[n]
        if k is not None:
            keys.add(k)
    return Stroke(keys)

steno_to_stroke.system = None


def replay(blackbox, name, test):
    # Hide from traceback on assertions (reduce output size for failed tests).
    __tracebackhide__ = operator.methodcaller('errisinstance', AssertionError)
    definitions, instructions = test.strip().rsplit('\n\n', 1)
    for steno, translation in ast.literal_eval(
        '{' + definitions + '}'
    ).items():
        blackbox.dictionary.set(normalize_steno(steno), translation)
    # Track line number for a more user-friendly assertion message.
    lines = test.split('\n')
    lnum = len(lines)-3 - test.rstrip().rsplit('\n\n', 1)[1].count('\n')
    for step in re.split('(?<=[^\\\\])\n', instructions):
        # Mark current instruction's line.
        lnum += 1
        step = step.strip()
        # Support for changing some settings on the fly.
        if step.startswith(':'):
            action = step[1:]
            if action == 'start_attached':
                blackbox.formatter.start_attached = True
            elif action == 'spaces_after':
                blackbox.formatter.set_space_placement('After Output')
            elif action == 'spaces_before':
                blackbox.formatter.set_space_placement('Before Output')
            else:
                raise ValueError('invalid action:\n%s' % msg)
            continue
        # Replay strokes.
        strokes, output = step.split(None, 1)
        for s in normalize_steno(strokes.strip()):
            blackbox.translator.translate(steno_to_stroke(s))
        # Check output.
        expected_output = ast.literal_eval(output.strip())
        msg = (
            name + '\n' +
            '\n'.join(('> ' if n == lnum else '  ') + l
                      for n, l in enumerate(lines)) + '\n' +
            '   ' + repr(blackbox.output.text) + '\n'
            '!= ' + repr(expected_output)
        )
        assert blackbox.output.text == expected_output, msg

def replay_doc(f):
    name = f.__name__
    test = textwrap.dedent(f.__doc__)
    # Use a lamdbda to reduce output size for failed tests.
    new_f = lambda bb: (f(bb), replay(bb, name, test))
    return functools.wraps(f)(new_f)


class BlackboxTester:

    @classmethod
    def setup_class(cls):
        for name in dir(cls):
            if name.startswith('test_'):
                setattr(cls, name, replay_doc(getattr(cls, name)))

    def setup_method(self):
        self.output = CaptureOutput()
        self.formatter = Formatter()
        self.formatter.set_output(self.output)
        self.translator = Translator()
        self.translator.set_min_undo_length(100)
        self.translator.add_listener(self.formatter.format)
        self.dictionary = self.translator.get_dictionary()
        self.dictionary.set_dicts([StenoDictionary()])



from functools import partial
import os
import tempfile

import pytest

from plover import system
from plover.config import Config, DictionaryConfig
from plover.engine import ErroredDictionary, StenoEngine
from plover.machine.base import StenotypeBase
from plover.machine.keymap import Keymap
from plover.registry import Registry
from plover.steno_dictionary import StenoDictionaryCollection

from .utils import make_dict


class FakeMachine(StenotypeBase):

    instance = None

    def __init__(self, options):
        super().__init__()
        self.options = options
        self.is_suppressed = False

    @classmethod
    def get_keys(cls):
        return system.KEYS

    def start_capture(self):
        assert FakeMachine.instance is None
        FakeMachine.instance = self
        self._initializing()
        self._ready()

    def stop_capture(self):
        FakeMachine.instance = None
        self._stopped()

    def set_suppression(self, enabled):
        self.is_suppressed = enabled

class FakeKeyboardEmulation:

    def send_backspaces(self, b):
        pass

    def send_string(self, s):
        pass

    def send_key_combination(self, c):
        pass

class FakeEngine(StenoEngine):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.events = []
        def hook_callback(hook, *args, **kwargs):
            self.events.append((hook, args, kwargs))
        for hook in self.HOOKS:
            self.hook_connect(hook, partial(hook_callback, hook))

    def _in_engine_thread(self):
        return True

    def quit(self, code=0):
        self._same_thread_hook(self._quit, code)

    def start(self):
        StenoEngine.start(self)


@pytest.fixture
def engine(monkeypatch):
    FakeMachine.instance = None
    registry = Registry()
    registry.update()
    registry.register_plugin('machine', 'Fake', FakeMachine)
    monkeypatch.setattr('plover.config.registry', registry)
    monkeypatch.setattr('plover.engine.registry', registry)
    cfg = Config()
    kbd = FakeKeyboardEmulation()
    cfg_file = tempfile.NamedTemporaryFile(prefix='plover',
                                           suffix='config',
                                           delete=False)
    try:
        cfg.target_file = cfg_file.name
        cfg['dictionaries'] = []
        cfg['machine_type'] = 'Fake'
        cfg['system_keymap'] = [(k, k) for k in system.KEYS]
        cfg.save(cfg_file)
        cfg_file.close()
        yield FakeEngine(cfg, kbd)
    finally:
        os.unlink(cfg_file.name)


def test_engine(engine):
    # Config load.
    assert engine.load_config()
    assert engine.events == []
    # Startup.
    engine.start()
    assert engine.events == [
        ('machine_state_changed', ('Fake', 'initializing'), {}),
        ('machine_state_changed', ('Fake', 'connected'), {}),
        ('config_changed', (engine.config,), {}),
    ]
    assert FakeMachine.instance is not None
    assert not FakeMachine.instance.is_suppressed
    # Output enabled.
    engine.events.clear()
    engine.output = True
    assert engine.events == [
        ('output_changed', (True,), {}),
    ]
    assert FakeMachine.instance.is_suppressed
    # Machine reconnection.
    engine.events.clear()
    engine.reset_machine()
    assert engine.events == [
        ('machine_state_changed', ('Fake', 'stopped'), {}),
        ('machine_state_changed', ('Fake', 'initializing'), {}),
        ('machine_state_changed', ('Fake', 'connected'), {}),
    ]
    assert FakeMachine.instance is not None
    assert FakeMachine.instance.is_suppressed
    # No machine reset on keymap change.
    engine.events.clear()
    new_keymap = Keymap(system.KEYS, system.KEYS)
    new_keymap.set_mappings(zip(system.KEYS, reversed(system.KEYS)))
    config_update = { 'system_keymap': new_keymap }
    assert FakeMachine.instance.keymap != new_keymap
    engine.config = config_update
    assert engine.events == [
        ('config_changed', (config_update,), {}),
    ]
    assert FakeMachine.instance.keymap == new_keymap
    # Output disabled
    engine.events.clear()
    engine.output = False
    assert engine.events == [
        ('output_changed', (False,), {}),
    ]
    assert not FakeMachine.instance.is_suppressed
    # Stopped.
    engine.events.clear()
    engine.quit(42)
    assert engine.join() == 42
    assert engine.events == [
        ('machine_state_changed', ('Fake', 'stopped'), {}),
        ('quit', (), {}),
    ]
    assert FakeMachine.instance is None

def test_loading_dictionaries(engine):
    def check_loaded_events(actual_events, expected_events):
        assert len(actual_events) == len(expected_events)
        for n, event in enumerate(actual_events):
            event_type, event_args, event_kwargs = event
            msg = 'event %u: %r' % (n, event)
            assert event_type == 'dictionaries_loaded', msg
            assert event_kwargs == {}, msg
            assert len(event_args) == 1, msg
            assert isinstance(event_args[0], StenoDictionaryCollection), msg
            assert [
                (d.path, d.enabled, isinstance(d, ErroredDictionary))
                for d in event_args[0].dicts
            ] == expected_events[n], msg
    with \
            make_dict(b'{}', 'json', 'valid1') as valid_dict_1, \
            make_dict(b'{}', 'json', 'valid2') as valid_dict_2, \
            make_dict(b'', 'json', 'invalid1') as invalid_dict_1, \
            make_dict(b'', 'json', 'invalid2') as invalid_dict_2:
        engine.start()
        for new_dictionaries, *expected_events in (
            # Load one valid dictionary.
            [[
                # path, enabled
                (valid_dict_1, True),
            ], [
                # path, enabled, errored
                (valid_dict_1, True, False),
            ]],
            # Load another invalid dictionary.
            [[
                (valid_dict_1, True),
                (invalid_dict_1, True),
            ], [
                (valid_dict_1, True, False),
                (invalid_dict_1, True, True),
            ]],
            # Disable first dictionary.
            [[
                (valid_dict_1, False),
                (invalid_dict_1, True),
            ], [
                (valid_dict_1, False, False),
                (invalid_dict_1, True, True),
            ]],
            # Replace invalid dictonary with another invalid one.
            [[
                (valid_dict_1, False),
                (invalid_dict_2, True),
            ], [
                (valid_dict_1, False, False),
            ], [
                (valid_dict_1, False, False),
                (invalid_dict_2, True, True),
            ]]
        ):
            engine.events.clear()
            config_update = {
                'dictionaries': [DictionaryConfig(*d)
                                 for d in new_dictionaries]
            }
            engine.config = dict(config_update)
            assert engine.events[0] == ('config_changed', (config_update,), {})
            check_loaded_events(engine.events[1:], expected_events)
        # Simulate an outdated dictionary.
        engine.events.clear()
        engine.dictionaries[valid_dict_1].timestamp -= 1
        engine.config = {}
        check_loaded_events(engine.events, [[
            (invalid_dict_2, True, True),
        ], [
            (valid_dict_1, False, False),
            (invalid_dict_2, True, True),
        ]])



from contextlib import contextmanager
import os
import tempfile

from plover.misc import normalize_path


@contextmanager
def make_dict(contents, extension=None, name=None):
    kwargs = { 'delete': False }
    if name is not None:
        kwargs['prefix'] = name + '_'
    if extension is not None:
        kwargs['suffix'] = '.' + extension
    tf = tempfile.NamedTemporaryFile(**kwargs)
    try:
        tf.write(contents)
        tf.close()
        yield normalize_path(tf.name)
    finally:
        os.unlink(tf.name)




# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

import io
import json
from collections import defaultdict


DICT_NAMES = ['main.json',
              'commands.json',
              'user.json']
DICT_PATH = 'plover/assets/'


def test_no_duplicates_categorized_files():
    d = defaultdict(list)
    dictionary = None
    def read_key_pairs(pairs):
        for key, value in pairs:
            d[key].append((value, dictionary))
        return d
    for dictionary in map(lambda x: DICT_PATH + x, DICT_NAMES):
        with io.open(dictionary, encoding='utf-8') as fp:
            json.load(fp, object_pairs_hook=read_key_pairs)
    msg_list = []
    has_duplicate = False
    for key, value_list in d.items():
        if len(value_list) > 1:
            has_duplicate = True
            msg_list.append('key: %s\n' % key)
            for value in value_list:
                msg_list.append('%r in %s\n' % value)
    msg = '\n' + ''.join(msg_list)
    assert not has_duplicate, msg



# -*- coding: utf-8 -*-

import biplist
import os.path

application = defines.get('app', './dist/Plover.app')
appname = os.path.basename(application)


def icon_from_app(app_path):
    plist_path = os.path.join(app_path, 'Contents', 'Info.plist')
    plist = biplist.readPlist(plist_path)
    icon_name = plist['CFBundleIconFile']
    icon_root, icon_ext = os.path.splitext(icon_name)
    if not icon_ext:
        icon_ext = '.icns'
    icon_name = icon_root + icon_ext
    return os.path.join(app_path, 'Contents', 'Resources', icon_name)

format = defines.get('format', 'UDBZ')

# Files to include
files = [application]

# Symlinks to create
symlinks = {'Applications': '/Applications'}

# Volume icon
badge_icon = icon_from_app(application)

# Where to put the icons
icon_locations = {
    appname:        (114, 244),
    'Applications': (527, 236)
}

background = 'osx/dmg_resources/background.png'

show_status_bar = False
show_tab_view = False
show_toolbar = False
show_pathbar = False
show_sidebar = False

# Window position in ((x, y), (w, h)) format
window_rect = ((100, 100), (640, 380))

default_view = 'icon-view'
show_icon_preview = False
include_icon_view_settings = 'auto'
include_list_view_settings = 'auto'

arrange_by = None
grid_offset = (0, 0)
grid_spacing = 50
scroll_position = (0, 0)
label_pos = 'bottom'
text_size = 16
icon_size = 86



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"""Unit tests for translation.py."""

from collections import namedtuple
import copy
import operator
import sys

from plover.steno_dictionary import StenoDictionary, StenoDictionaryCollection
from plover.translation import Translation, Translator, _State
from plover.translation import escape_translation, unescape_translation
from plover.steno import Stroke, normalize_steno

from plover_build_utils.testing import steno_to_stroke as stroke

from . import parametrize


if sys.platform.startswith('darwin'):
    BACK_STRING = '{#Alt_L(BackSpace)}{^}'
else:
    BACK_STRING = '{#Control_L(BackSpace)}{^}'


def test_no_translation():
    t = Translation([stroke('S'), stroke('T')], None)
    assert t.strokes == [stroke('S'), stroke('T')]
    assert t.rtfcre == ('S', 'T')
    assert t.english is None

def test_translation():
    t = Translation([stroke('S'), stroke('T')], 'translation')
    assert t.strokes == [stroke('S'), stroke('T')]
    assert t.rtfcre == ('S', 'T')
    assert t.english == 'translation'


class TestTranslatorStateSize:

    class FakeState(_State):

        def __init__(self):
            _State.__init__(self)
            self.restrict_calls = []

        def restrict_size(self, n):
            self.restrict_calls.append(n)

    def _check_size_call(self, size):
        assert self.s.restrict_calls[-1] == size

    def _check_no_size_call(self):
        assert self.s.restrict_calls == []

    def clear(self):
        self.s.restrict_calls = []

    def setup_method(self):
        self.t = Translator()
        self.s = type(self).FakeState()
        self.t._state = self.s
        self.d = StenoDictionary()
        self.dc = StenoDictionaryCollection([self.d])
        self.t.set_dictionary(self.dc)

    def test_dictionary_update_grows_size1(self):
        self.d[('S',)] = '1'
        self._check_size_call(1)

    def test_dictionary_update_grows_size4(self):
        self.d[('S', 'PT', '-Z', 'TOP')] = 'hi'
        self._check_size_call(4)

    def test_dictionary_update_no_grow(self):
        self.t.set_min_undo_length(4)
        self._check_size_call(4)
        self.clear()
        self.d[('S', 'T')] = 'nothing'
        self._check_size_call(4)

    def test_dictionary_update_shrink(self):
        self.d[('S', 'T', 'P', '-Z', '-D')] = '1'
        self._check_size_call(5)
        self.clear()
        self.d[('A', 'P')] = '2'
        self._check_no_size_call()
        del self.d[('S', 'T', 'P', '-Z', '-D')]
        self._check_size_call(2)

    def test_dictionary_update_no_shrink(self):
        self.t.set_min_undo_length(7)
        self.d[('S', 'T', 'P', '-Z', '-D')] = '1'
        del self.d[('S', 'T', 'P', '-Z', '-D')]
        self._check_size_call(7)

    def test_translation_calls_restrict(self):
        self.t.translate(stroke('S'))
        self._check_size_call(0)


def test_listeners():
    output1 = []
    def listener1(undo, do, prev):
        output1.append((undo, do, prev))

    output2 = []
    def listener2(undo, do, prev):
        output2.append((undo, do, prev))

    t = Translator()
    s = stroke('S')
    tr = Translation([s], None)
    expected_output = [([], [tr], [tr])]

    t.translate(s)

    t.add_listener(listener1)
    t.translate(s)
    assert output1 == expected_output

    del output1[:]
    t.add_listener(listener2)
    t.translate(s)
    assert output1 == expected_output
    assert output2 == expected_output

    del output1[:]
    del output2[:]
    t.add_listener(listener2)
    t.translate(s)
    assert output1 == expected_output
    assert output2 == expected_output

    del output1[:]
    del output2[:]
    t.remove_listener(listener1)
    t.translate(s)
    assert output1 == []
    assert output2 == expected_output

    del output1[:]
    del output2[:]
    t.remove_listener(listener2)
    t.translate(s)
    assert output1 == []
    assert output2 == []


def test_changing_state():
    output = []
    def listener(undo, do, prev):
        prev = list(prev) if prev else None
        output.append((undo, do, prev))

    d = StenoDictionary()
    d[('S', 'P')] = 'hi'
    dc = StenoDictionaryCollection([d])
    t = Translator()
    t.set_dictionary(dc)
    t.translate(stroke('T'))
    t.translate(stroke('S'))
    s = copy.deepcopy(t.get_state())

    t.add_listener(listener)

    expected = [([Translation([stroke('S')], None)],
                 [Translation([stroke('S'), stroke('P')], 'hi')],
                 [Translation([stroke('T')], None)])]
    t.translate(stroke('P'))
    assert output == expected

    del output[:]
    t.set_state(s)
    t.translate(stroke('P'))
    assert output == expected

    del output[:]
    t.clear_state()
    t.translate(stroke('P'))
    assert output == [([], [Translation([stroke('P')], None)], None)]

    del output[:]
    t.set_state(s)
    t.translate(stroke('P'))
    assert output == [([],
                       [Translation([stroke('P')], None)],
                       [Translation([stroke('S'), stroke('P')], 'hi')])]


def test_translator():

    # It's not clear that this test is needed anymore. There are separate
    # tests for _translate_stroke and test_translate_calls_translate_stroke
    # makes sure that translate calls it properly. But since I already wrote
    # this test I'm going to keep it.

    class Output:
        def __init__(self):
            self._output = []

        def write(self, undo, do, prev):
            for t in undo:
                self._output.pop()
            for t in do:
                if t.english:
                    self._output.append(t.english)
                else:
                    self._output.append('/'.join(t.rtfcre))

        def get(self):
            return ' '.join(self._output)

        def clear(self):
            del self._output[:]

    d = StenoDictionary()
    out = Output()
    t = Translator()
    dc = StenoDictionaryCollection([d])
    t.set_dictionary(dc)
    t.add_listener(out.write)

    t.translate(stroke('S'))
    assert out.get() == 'S'
    t.translate(stroke('T'))
    assert out.get() == 'S T'
    t.translate(stroke('*'))
    assert out.get() == 'S'
    t.translate(stroke('*'))
    # Undo buffer ran out
    assert out.get() == 'S ' + BACK_STRING

    t.set_min_undo_length(3)
    out.clear()
    t.translate(stroke('S'))
    assert out.get() == 'S'
    t.translate(stroke('T'))
    assert out.get() == 'S T'
    t.translate(stroke('*'))
    assert out.get() == 'S'
    t.translate(stroke('*'))
    assert out.get() == ''

    out.clear()
    d[('S',)] = 't1'
    d[('T',)] = 't2'
    d[('S', 'T')] = 't3'

    t.translate(stroke('S'))
    assert out.get() == 't1'
    t.translate(stroke('T'))
    assert out.get() == 't3'
    t.translate(stroke('T'))
    assert out.get() == 't3 t2'
    t.translate(stroke('S'))
    assert out.get() == 't3 t2 t1'
    t.translate(stroke('*'))
    assert out.get() == 't3 t2'
    t.translate(stroke('*'))
    assert out.get() == 't3'
    t.translate(stroke('*'))
    assert out.get() == 't1'
    t.translate(stroke('*'))
    assert out.get() == ''

    t.translate(stroke('S'))
    assert out.get() == 't1'
    t.translate(stroke('T'))
    assert out.get() == 't3'
    t.translate(stroke('T'))
    assert out.get() == 't3 t2'

    d[('S', 'T', 'T')] = 't4'
    d[('S', 'T', 'T', 'S')] = 't5'

    t.translate(stroke('S'))
    assert out.get() == 't5'
    t.translate(stroke('*'))
    assert out.get() == 't3 t2'
    t.translate(stroke('*'))
    assert out.get() == 't3'
    t.translate(stroke('T'))
    assert out.get() == 't4'
    t.translate(stroke('S'))
    assert out.get() == 't5'
    t.translate(stroke('S'))
    assert out.get() == 't5 t1'
    t.translate(stroke('*'))
    assert out.get() == 't5'
    t.translate(stroke('*'))
    assert out.get() == 't4'
    t.translate(stroke('*'))
    assert out.get() == 't3'
    t.translate(stroke('*'))
    assert out.get() == 't1'
    t.translate(stroke('*'))
    assert out.get() == ''

    d.clear()

    s = stroke('S')
    t.translate(s)
    t.translate(s)
    t.translate(s)
    t.translate(s)
    s = stroke('*')
    t.translate(s)
    t.translate(s)
    t.translate(s)
    t.translate(s)
    # Not enough undo to clear output.
    assert out.get() == 'S ' + BACK_STRING

    out.clear()
    t.remove_listener(out.write)
    t.translate(stroke('S'))
    assert out.get() == ''


class TestState:

    def setup_method(self):
        self.a = Translation([stroke('S')], None)
        self.b = Translation([stroke('T'), stroke('-D')], None)
        self.c = Translation([stroke('-Z'), stroke('P'), stroke('T*')], None)

    def test_prev_list0(self):
        s = _State()
        assert s.prev() is None

    def test_prev_list1(self):
        s = _State()
        s.translations = [self.a]
        assert s.prev() == [self.a]

    def test_prev_list2(self):
        s = _State()
        s.translations = [self.a, self.b]
        assert s.prev() == [self.a, self.b]

    def test_prev_tail1(self):
        s = _State()
        s.translations = [self.a]
        s.tail = self.b
        assert s.prev() == [self.a]

    def test_prev_tail0(self):
        s = _State()
        s.tail = self.b
        assert s.prev() == [self.b]

    def test_restrict_size_zero_on_empty(self):
        s = _State()
        s.restrict_size(0)
        assert s.translations == []
        assert s.tail is None

    def test_restrict_size_zero_on_one_stroke(self):
        s = _State()
        s.translations = [self.a]
        s.restrict_size(0)
        assert s.translations == [self.a]
        assert s.tail is None

    def test_restrict_size_to_exactly_one_stroke(self):
        s = _State()
        s.translations = [self.a]
        s.restrict_size(1)
        assert s.translations == [self.a]
        assert s.tail is None

    def test_restrict_size_to_one_on_two_strokes(self):
        s = _State()
        s.translations = [self.b]
        s.restrict_size(1)
        assert s.translations == [self.b]
        assert s.tail is None

    def test_restrict_size_to_one_on_two_translations(self):
        s = _State()
        s.translations = [self.b, self.a]
        s.restrict_size(1)
        assert s.translations == [self.a]
        assert s.tail == self.b

    def test_restrict_size_to_one_on_two_translations_too_big(self):
        s = _State()
        s.translations = [self.a, self.b]
        s.restrict_size(1)
        assert s.translations == [self.b]
        assert s.tail == self.a

    def test_restrict_size_lose_translations(self):
        s = _State()
        s.translations = [self.a, self.b, self.c]
        s.restrict_size(2)
        assert s.translations == [self.c]
        assert s.tail == self.b

    def test_restrict_size_multiple_translations(self):
        s = _State()
        s.translations = [self.a, self.b, self.c]
        s.restrict_size(5)
        assert s.translations == [self.b, self.c]
        assert s.tail == self.a


class TestTranslateStroke:

    class CaptureOutput:
        output = namedtuple('output', 'undo do prev')

        def __init__(self):
            self.output = []

        def __call__(self, undo, new, prev):
            prev = list(prev) if prev else None
            self.output = type(self).output(undo, new, prev)

    def t(self, strokes):
        """A quick way to make a translation."""
        strokes = [stroke(x) for x in strokes.split('/')]
        key = tuple(s.rtfcre for s in strokes)
        translation = self.dc.lookup(key)
        return Translation(strokes, translation)

    def lt(self, translations):
        """A quick way to make a list of translations."""
        return [self.t(x) for x in translations.split()]

    def define(self, key, value):
        key = normalize_steno(key)
        self.d[key] = value

    def translate(self, steno):
        self.tlor.translate(stroke(steno))

    def _check_translations(self, expected):
        # Hide from traceback on assertions (reduce output size for failed tests).
        __tracebackhide__ = operator.methodcaller('errisinstance', AssertionError)
        msg = '''
        translations:
            results: %s
            expected: %s
        ''' % (self.s.translations, expected)
        assert self.s.translations == expected, msg

    def _check_output(self, undo, do, prev):
        # Hide from traceback on assertions (reduce output size for failed tests).
        __tracebackhide__ = operator.methodcaller('errisinstance', AssertionError)
        msg = '''
        output:
            results: -%s
                     +%s
                     [%s]
            expected: -%s
                      +%s
                      [%s]
        ''' % (self.o.output + (undo, do, prev))
        assert self.o.output == (undo, do, prev), msg

    def setup_method(self):
        self.d = StenoDictionary()
        self.dc = StenoDictionaryCollection([self.d])
        self.s = _State()
        self.o = self.CaptureOutput()
        self.tlor = Translator()
        self.tlor.set_dictionary(self.dc)
        self.tlor.add_listener(self.o)
        self.tlor.set_state(self.s)

    def test_first_stroke(self):
        self.translate('-B')
        self._check_translations(self.lt('-B'))
        self._check_output([], self.lt('-B'), None)

    def test_second_stroke(self):
        self.define('S/P', 'spiders')
        self.s.translations = self.lt('S')
        self.translate('-T')
        self._check_translations(self.lt('S -T'))
        self._check_output([], self.lt('-T'), self.lt('S'))

    def test_second_stroke_tail(self):
        self.s.tail = self.t('T/A/EU/L')
        self.translate('-E')
        self._check_translations(self.lt('E'))
        self._check_output([], self.lt('E'), self.lt('T/A/EU/L'))

    def test_with_translation_1(self):
        self.define('S', 'is')
        self.define('-T', 'that')
        self.s.translations = self.lt('S')
        self.tlor.set_min_undo_length(2)
        self.translate('-T')
        self._check_translations(self.lt('S -T'))
        self._check_output([], self.lt('-T'), self.lt('S'))
        assert self.o.output.do[0].english == 'that'

    def test_with_translation_2(self):
        self.define('S', 'is')
        self.define('-T', 'that')
        self.s.translations = self.lt('S')
        self.tlor.set_min_undo_length(1)
        self.translate('-T')
        self._check_translations(self.lt('-T'))
        self._check_output([], self.lt('-T'), self.lt('S'))
        assert self.o.output.do[0].english == 'that'

    def test_finish_two_translation(self):
        self.define('S/T', 'hello')
        self.s.translations = self.lt('S')
        self.translate('T')
        self._check_translations(self.lt('S/T'))
        self._check_output(self.lt('S'), self.lt('S/T'), None)
        assert self.o.output.do[0].english == 'hello'
        assert self.o.output.do[0].replaced == self.lt('S')

    def test_finish_three_translation(self):
        self.define('S/T/-B', 'bye')
        self.s.translations = self.lt('S T')
        self.translate('-B')
        self._check_translations(self.lt('S/T/-B'))
        self._check_output(self.lt('S T'), self.lt('S/T/-B'), None)
        assert self.o.output.do[0].english == 'bye'
        assert self.o.output.do[0].replaced == self.lt('S T')

    def test_replace_translation(self):
        self.define('S/T/-B', 'longer')
        self.s.translations = self.lt('S/T')
        self.translate('-B')
        self._check_translations(self.lt('S/T/-B'))
        self._check_output(self.lt('S/T'), self.lt('S/T/-B'), None)
        assert self.o.output.do[0].english == 'longer'
        assert self.o.output.do[0].replaced == self.lt('S/T')

    def test_undo(self):
        self.s.translations = self.lt('POP')
        self.translate('*')
        self._check_translations([])
        self._check_output(self.lt('POP'), [], None)

    def test_empty_undo(self):
        self.translate('*')
        self._check_translations([])
        self._check_output([], [Translation([Stroke('*')], BACK_STRING)], None)

    def test_undo_translation(self):
        self.define('P/P', 'pop')
        self.translate('P')
        self.translate('P')
        self.translate('*')
        self._check_translations(self.lt('P'))
        self._check_output(self.lt('P/P'), self.lt('P'), None)

    def test_undo_longer_translation(self):
        self.define('P/P/-D', 'popped')
        self.translate('P')
        self.translate('P')
        self.translate('-D')
        self.translate('*')
        self._check_translations(self.lt('P P'))
        self._check_output(self.lt('P/P/-D'), self.lt('P P'), None)

    def test_undo_tail(self):
        self.s.tail = self.t('T/A/EU/L')
        self.translate('*')
        self._check_translations([])
        self._check_output([], [Translation([Stroke('*')], BACK_STRING)], [self.s.tail])

    def test_suffix_folding(self):
        self.define('K-L', 'look')
        self.define('-G', '{^ing}')
        lt = self.lt('K-LG')
        lt[0].english = 'look {^ing}'
        self.translate('K-LG')
        self._check_translations(lt)

    def test_suffix_folding_multi_stroke(self):
        self.define('E/HR', 'he will')
        self.define('-S', '{^s}')
        self.translate('E')
        self.translate('HR-S')
        output = ' '.join(t.english for t in self.s.translations)
        assert output == 'he will {^s}'

    def test_suffix_folding_doesnt_interfere(self):
        self.define('E/HR', 'he will')
        self.define('-S', '{^s}')
        self.define('E', 'he')
        self.define('HR-S', 'also')
        self.translate('E')
        self.translate('HR-S')
        output = ' '.join(t.english for t in self.s.translations)
        assert output == 'he also'

    def test_suffix_folding_no_suffix(self):
        self.define('K-L', 'look')
        lt = self.lt('K-LG')
        assert lt[0].english is None
        self.translate('K-LG')
        self._check_translations(lt)

    def test_suffix_folding_no_main(self):
        self.define('-G', '{^ing}')
        lt = self.lt('K-LG')
        assert lt[0].english is None
        self.translate('K-LG')
        self._check_translations(lt)

    def test_retrospective_insert_space(self):
        self.define('T/E/S/T', 'a longer key')
        self.define('PER', 'perfect')
        self.define('SWAEUGS', 'situation')
        self.define('PER/SWAEUGS', 'persuasion')
        self.define('SP*', '{*?}')
        self.translate('PER')
        self.translate('SWAEUGS')
        self.translate('SP*')
        lt = self.lt('PER')
        undo = self.lt('PER/SWAEUGS')
        undo[0].replaced = lt
        do = self.lt('SP*')
        do[0].english = 'perfect situation'
        do[0].is_retrospective_command = True
        do[0].replaced = undo
        self._check_translations(do)
        self._check_output(undo, do, None)

    def test_retrospective_insert_space_undefined(self):
        # Should work when beginning or ending strokes aren't defined
        self.define('T/E/S/T', 'a longer key')
        self.define('STWR/STWR', 'test')
        self.define('SP*', '{*?}')
        self.translate('STWR')
        self.translate('STWR')
        self.translate('SP*')
        lt = self.lt('STWR')
        undo = self.lt('STWR/STWR')
        undo[0].replaced = lt
        do = self.lt('SP*')
        do[0].english = 'STWR STWR'
        do[0].is_retrospective_command = True
        do[0].replaced = undo
        self._check_translations(do)
        self._check_output(undo, do, None)

    def test_retrospective_delete_space(self):
        self.define('T/E/S/T', 'a longer key')
        self.define('K', 'kick')
        self.define('PW', 'back')
        self.define('SP*', '{*!}')
        self.translate('K')
        self.translate('PW')
        self.translate('SP*')
        undo = self.lt('K PW')
        do = self.lt('SP*')
        do[0].english = 'kick{^~|^}back'
        do[0].is_retrospective_command = True
        do[0].replaced = undo
        self._check_translations(do)
        self._check_output(undo, do, None)

    def test_retrospective_delete_space_with_number(self):
        self.define('T/E/S/T', 'a longer key')
        self.define('U', 'user')
        self.define('SP*', '{*!}')
        self.translate('U')
        self.translate('1-')
        self.translate('SP*')
        undo = self.lt('U 1-')
        do = self.lt('SP*')
        do[0].english = 'user{^~|^}{&1}'
        do[0].is_retrospective_command = True
        do[0].replaced = undo
        self._check_translations(do)
        self._check_output(undo, do, None)

    def test_retrospective_delete_space_with_period(self):
        self.define('T/E/S/T', 'a longer key')
        self.define('P-P', '{.}')
        self.define('SH*', 'zshrc')
        self.define('SP*', '{*!}')
        self.translate('P-P')
        self.translate('SH*')
        self.translate('SP*')
        undo = self.lt('P-P SH*')
        do = self.lt('SP*')
        do[0].english = '{.}{^~|^}zshrc'
        do[0].is_retrospective_command = True
        do[0].replaced = undo
        self._check_translations(do)
        self._check_output(undo, do, None)

    def test_retrospective_toggle_asterisk(self):
        self.define('T/E/S/T', 'a longer key')
        self.define('S', 'see')
        self.define('S*', 'sea')
        self.define('A*', '{*}')
        self.translate('S')
        self.translate('A*')
        self._check_translations(self.lt('S*'))
        self._check_output(self.lt('S'), self.lt('S*'), None)

    def test_retrospective_toggle_empty(self):
        self.define('A*', '{*}')
        self.translate('A*')
        self._check_translations(self.lt(''))
        assert self.o.output == []

    def test_retrospective_toggle_asterisk_replaced1(self):
        self.define('P-P', '{.}')
        self.define('SKEL', 'cancel')
        self.define('SKEL/TO-PB', 'skeleton')
        self.define('SKEL/TO*PB', 'not skeleton!')
        self.define('A*', '{*}')
        self.translate('P-P')
        self.translate('SKEL')
        self.translate('TO-PB')
        self.translate('A*')
        self._check_translations(self.lt('SKEL/TO*PB'))
        self._check_output(self.lt('SKEL/TO-PB'),
                          self.lt('SKEL/TO*PB'),
                          self.lt('P-P'))

    def test_retrospective_toggle_asterisk_replaced2(self):
        self.define('P-P', '{.}')
        self.define('SKEL', 'cancel')
        self.define('SKEL/TO-PB', 'skeleton')
        self.define('TO*PB', '{^ton}')
        self.define('A*', '{*}')
        self.translate('P-P')
        self.translate('SKEL')
        self.translate('TO-PB')
        self.translate('A*')
        self._check_translations(self.lt('SKEL TO*PB'))
        self._check_output(self.lt('SKEL/TO-PB'),
                          self.lt('SKEL TO*PB'),
                          self.lt('P-P'))

    def test_repeat_last_stroke1(self):
        self.define('T/E/S/T', 'a longer key')
        self.define('TH', 'this')
        self.define('R*', '{*+}')
        self.translate('TH')
        self.translate('R*')
        undo = []
        do = self.lt('TH')
        state = self.lt('TH TH')
        self._check_translations(state)
        self._check_output(undo, do, do)

    def test_repeat_last_stroke2(self):
        self.define('T/E/S/T', 'a longer key')
        self.define('THA', 'that')
        self.define('R*', '{*+}')
        self.translate('THA')
        self.translate('R*')
        undo = []
        do = self.lt('THA')
        state = self.lt('THA THA')
        self._check_translations(state)
        self._check_output(undo, do, do)

    def test_untranslate_translation(self):
        self.tlor.set_min_undo_length(2)
        self.define('TH', 'this')
        self.define('THA', 'that')
        self.translate('TH')
        self.translate('THA')
        self.tlor.untranslate_translation(self.t('THA'))
        self.tlor.untranslate_translation(self.t('TH'))
        self.tlor.flush()
        self._check_output(self.lt('TH THA'), [], None)


ESCAPE_UNESCAPE_TRANSLATION_TESTS = (
    # No change.
    lambda: ('foobar', 'foobar'),
    lambda: (r'\\', r'\\'),
    lambda: ('\\\\\\', '\\\\\\'), # -> \\\
    # Basic support: \n, \r, \t.
    lambda: ('\n', r'\n'),
    lambda: ('\r', r'\r'),
    lambda: ('\t', r'\t'),
    # Allow a literal \n, \r, or \t by doubling the \.
    lambda: (r'\n', r'\\n'),
    lambda: (r'\r', r'\\r'),
    lambda: (r'\t', r'\\t'),
    lambda: (r'\\n', r'\\\n'),
    lambda: (r'\\r', r'\\\r'),
    lambda: (r'\\t', r'\\\t'),
    # A little more complex.
    lambda: ('\tfoo\nbar\r', r'\tfoo\nbar\r'),
    lambda: ('\\tfoo\\nbar\\r', r'\\tfoo\\nbar\\r'),
)

@parametrize(ESCAPE_UNESCAPE_TRANSLATION_TESTS)
def test_escape_unescape_translation(raw, escaped):
    assert unescape_translation(escaped) == raw
    assert escape_translation(raw) == escaped



#!/usr/bin/env python3

import os
import subprocess
import sys

from .download import download
from .install_wheels import WHEELS_CACHE, install_wheels


def get_pip(args=None):
    # Download `get-pip.py`.
    script = download('https://bootstrap.pypa.io/get-pip.py',
                      '35ac67d8b15c308b22e3234c8a6f44a652f2d869')
    # Make sure wheels cache directory exists to avoid warning.
    if not os.path.exists(WHEELS_CACHE):
        os.makedirs(WHEELS_CACHE)
    # Install pip/wheel...
    get_pip_cmd = [sys.executable, script, '-f', WHEELS_CACHE]
    if args is not None:
        get_pip_cmd.extend(args)
    subprocess.check_call(get_pip_cmd)
    # ...and cache them for the next iteration (if possible).
    try:
        import wheel
    except ImportError:
        pass
    else:
        install_wheels(['--no-install', 'pip', 'wheel'])


if __name__ == '__main__':
    get_pip(sys.argv[1:])



import objc
NSUserNotification = objc.lookUpClass('NSUserNotification')
NSUserNotificationCenter = objc.lookUpClass('NSUserNotificationCenter')
NSObject = objc.lookUpClass('NSObject')

from plover import log, __name__ as __software_name__
import logging


class OSXNotificationHandler(logging.Handler):
    """ Handler using OS X Notification Center to show messages. """

    def __init__(self):
        super().__init__()
        self.setLevel(log.WARNING)
        self.setFormatter(log.NoExceptionTracebackFormatter('%(message)s'))

    def emit(self, record):
        # Notification Center has no levels or timeouts.
        notification = NSUserNotification.alloc().init()

        notification.setTitle_(__software_name__.capitalize())
        notification.setSubtitle_(record.levelname.title())
        notification.setInformativeText_(self.format(record))

        ns = NSUserNotificationCenter.defaultUserNotificationCenter()
        ns.setDelegate_(always_present_delegator)
        ns.deliverNotification_(notification)


class AlwaysPresentNSDelegator(NSObject):
    """
    Custom delegator to force presenting even if Plover is in the foreground.
    """
    def userNotificationCenter_didActivateNotification_(self, ns, note):
        # Do nothing
        return

    def userNotificationCenter_shouldPresentNotification_(self, ns, note):
        # Force notification, even if frontmost application.
        return True

always_present_delegator = AlwaysPresentNSDelegator.alloc().init()



import signal

from PyQt5.QtCore import (
    QCoreApplication,
    QLibraryInfo,
    QTimer,
    QTranslator,
    Qt,
)
from PyQt5.QtWidgets import QApplication, QMessageBox

from plover import log
from plover import __name__ as __software_name__
from plover import __version__
from plover.oslayer.keyboardcontrol import KeyboardEmulation

from plover.gui_qt.engine import Engine
from plover.gui_qt.i18n import get_language, install_gettext


class Application:

    def __init__(self, config, use_qt_notifications):

        # This is done dynamically so localization
        # support can be configure beforehand.
        from plover.gui_qt.main_window import MainWindow

        self._app = None
        self._win = None
        self._engine = None
        self._translator = None

        QCoreApplication.setApplicationName(__software_name__.capitalize())
        QCoreApplication.setApplicationVersion(__version__)
        QCoreApplication.setOrganizationName('Open Steno Project')
        QCoreApplication.setOrganizationDomain('openstenoproject.org')

        self._app = QApplication([])
        self._app.setAttribute(Qt.AA_UseHighDpiPixmaps)

        # Enable localization of standard Qt controls.
        self._translator = QTranslator()
        translations_dir = QLibraryInfo.location(QLibraryInfo.TranslationsPath)
        self._translator.load('qtbase_' + get_language(), translations_dir)
        self._app.installTranslator(self._translator)

        QApplication.setQuitOnLastWindowClosed(False)

        self._app.engine = self._engine = Engine(config, KeyboardEmulation())
        # On macOS, quitting through the dock will result
        # in a direct call to `QCoreApplication.quit`.
        self._app.aboutToQuit.connect(self._app.engine.quit)

        signal.signal(signal.SIGINT, lambda signum, stack: self._engine.quit())

        # Make sure the Python interpreter runs at least every second,
        # so signals have a chance to be processed.
        self._timer = QTimer()
        self._timer.timeout.connect(lambda: None)
        self._timer.start(1000)

        self._win = MainWindow(self._engine, use_qt_notifications)

    def __del__(self):
        del self._win
        del self._app
        del self._engine
        del self._translator

    def run(self):
        self._app.exec_()
        return self._engine.join()


def show_error(title, message):
    print('%s: %s' % (title, message))
    app = QApplication([])
    QMessageBox.critical(None, title, message)
    del app


def main(config):
    # Setup internationalization support.
    install_gettext()
    use_qt_notifications = not log.has_platform_handler()
    app = Application(config, use_qt_notifications)
    code = app.run()
    del app
    return code



# Copyright (c) 2010-2011 Joshua Harlan Lifton.
# See LICENSE.txt for details.

# TODO: add tests for all machines
# TODO: add tests for new status callbacks

"""Base classes for machine types. Do not use directly."""

import binascii
import threading

import serial

from plover import log
from plover.machine.keymap import Keymap
from plover.misc import boolean


STATE_STOPPED = 'stopped'
STATE_INITIALIZING = 'initializing'
STATE_RUNNING = 'connected'
STATE_ERROR = 'disconnected'


class StenotypeBase:
    """The base class for all Stenotype classes."""

    # Layout of physical keys.
    KEYS_LAYOUT = ''
    # And special actions to map to.
    ACTIONS = ()
    # Fallback to use as machine type for finding a compatible keymap
    # if one is not already available for this machine type.
    KEYMAP_MACHINE_TYPE = None

    def __init__(self):
        # Setup default keymap with no translation of keys.
        keys = self.get_keys()
        self.keymap = Keymap(keys, keys)
        self.keymap.set_mappings(zip(keys, keys))
        self.stroke_subscribers = []
        self.state_subscribers = []
        self.state = STATE_STOPPED

    def set_keymap(self, keymap):
        """Setup machine keymap."""
        self.keymap = keymap

    def start_capture(self):
        """Begin listening for output from the stenotype machine."""
        pass

    def stop_capture(self):
        """Stop listening for output from the stenotype machine."""
        pass

    def add_stroke_callback(self, callback):
        """Subscribe to output from the stenotype machine.

        Argument:

        callback -- The function to call whenever there is output from
        the stenotype machine and output is being captured.

        """
        self.stroke_subscribers.append(callback)

    def remove_stroke_callback(self, callback):
        """Unsubscribe from output from the stenotype machine.

        Argument:

        callback -- A function that was previously subscribed.

        """
        self.stroke_subscribers.remove(callback)

    def add_state_callback(self, callback):
        self.state_subscribers.append(callback)
        
    def remove_state_callback(self, callback):
        self.state_subscribers.remove(callback)

    def _notify(self, steno_keys):
        """Invoke the callback of each subscriber with the given argument."""
        for callback in self.stroke_subscribers:
            callback(steno_keys)

    def set_suppression(self, enabled):
        '''Enable keyboard suppression.

        This is only of use for the keyboard machine,
        to suppress the keyboard when then engine is running.
        '''
        pass

    def suppress_last_stroke(self, send_backspaces):
        '''Suppress the last stroke key events after the fact.

        This is only of use for the keyboard machine,
        and the engine is resumed with a command stroke.

        Argument:

        send_backspaces -- The function to use to send backspaces.
        '''
        pass

    def _set_state(self, state):
        self.state = state
        for callback in self.state_subscribers:
            callback(state)

    def _stopped(self):
        self._set_state(STATE_STOPPED)

    def _initializing(self):
        self._set_state(STATE_INITIALIZING)

    def _ready(self):
        self._set_state(STATE_RUNNING)
            
    def _error(self):
        self._set_state(STATE_ERROR)

    @classmethod
    def get_actions(cls):
        """List of supported actions to map to."""
        return cls.ACTIONS

    @classmethod
    def get_keys(cls):
        return tuple(cls.KEYS_LAYOUT.split())

    @classmethod
    def get_option_info(cls):
        """Get the default options for this machine."""
        return {}


class ThreadedStenotypeBase(StenotypeBase, threading.Thread):
    """Base class for thread based machines.
    
    Subclasses should override run.
    """
    def __init__(self):
        threading.Thread.__init__(self)
        self.name += '-machine'
        StenotypeBase.__init__(self)
        self.finished = threading.Event()

    def run(self):
        """This method should be overridden by a subclass."""
        pass

    def start_capture(self):
        """Begin listening for output from the stenotype machine."""
        self.finished.clear()
        self._initializing()
        self.start()

    def stop_capture(self):
        """Stop listening for output from the stenotype machine."""
        self.finished.set()
        try:
            self.join()
        except RuntimeError:
            pass
        self._stopped()

class SerialStenotypeBase(ThreadedStenotypeBase):
    """For use with stenotype machines that connect via serial port.

    This class implements the three methods necessary for a standard
    stenotype interface: start_capture, stop_capture, and
    add_callback.

    """

    # Default serial parameters.
    SERIAL_PARAMS = {
        'port': None,
        'baudrate': 9600,
        'bytesize': 8,
        'parity': 'N',
        'stopbits': 1,
        'timeout': 2.0,
    }

    def __init__(self, serial_params):
        """Monitor the stenotype over a serial port.

        Keyword arguments are the same as the keyword arguments for a
        serial.Serial object.

        """
        ThreadedStenotypeBase.__init__(self)
        self.serial_port = None
        self.serial_params = serial_params

    def _close_port(self):
        if self.serial_port is None:
            return
        self.serial_port.close()
        self.serial_port = None

    def start_capture(self):
        self._close_port()

        try:
            self.serial_port = serial.Serial(**self.serial_params)
        except (serial.SerialException, OSError):
            log.warning('Can\'t open serial port', exc_info=True)
            self._error()
            return

        if not self.serial_port.isOpen():
            log.warning('Serial port is not open: %s', self.serial_params.get('port'))
            self._error()
            return

        return ThreadedStenotypeBase.start_capture(self)

    def stop_capture(self):
        """Stop listening for output from the stenotype machine."""
        ThreadedStenotypeBase.stop_capture(self)
        self._close_port()

    @classmethod
    def get_option_info(cls):
        """Get the default options for this machine."""
        sb = lambda s: int(float(s)) if float(s).is_integer() else float(s)
        converters = {
            'port': str,
            'baudrate': int,
            'bytesize': int,
            'parity': str,
            'stopbits': sb,
            'timeout': float,
            'xonxoff': boolean,
            'rtscts': boolean,
        }
        return {
            setting: (default, converters[setting])
            for setting, default in cls.SERIAL_PARAMS.items()
        }

    def _iter_packets(self, packet_size):
        """Yield packets of <packets_size> bytes until the machine is stopped.

        N.B.: to workaround the fact that the Toshiba Bluetooth stack
        on Windows does not correctly handle the read timeout setting
        (returning immediately if some data is already available):
        - the effective timeout is re-configured to <timeout/packet_size>
        - multiple reads are  done (until a packet is complete)
        - an incomplete packet will only be discarded if one of
          those reads return no data (but not on short read)
        """
        self.serial_port.timeout = max(
            self.serial_params.get('timeout', 1.0) / packet_size,
            0.01,
        )
        packet = b''
        while not self.finished.isSet():
            raw = self.serial_port.read(packet_size - len(packet))
            if not raw:
                if packet:
                    log.error('discarding incomplete packet: %s',
                              binascii.hexlify(packet))
                packet = b''
                continue
            packet += raw
            if len(packet) != packet_size:
                continue
            yield packet
            packet = b''



import os
import sys
import locale
import gettext

import pkg_resources

from plover import log


# Mark some strings for localization.
def _unused():
    # Machines.
    _('Keyboard')
    # States.
    _('stopped')
    _('initializing')
    _('connected')
    _('disconnected')


def get_language():
    env_vars = ['LANGUAGE']
    if sys.platform.startswith('linux'):
        env_vars.extend(('LC_ALL', 'LC_MESSAGES', 'LANG'))
    for var in env_vars:
        lang = os.environ.get(var)
        if lang is not None:
            return lang
    if sys.platform.startswith('linux'):
        lang, enc = locale.getdefaultlocale()
    elif sys.platform.startswith('darwin'):
        from AppKit import NSLocale
        lang_list = NSLocale.preferredLanguages()
        lang = lang_list[0] if lang_list else None
    elif sys.platform.startswith('win'):
        from ctypes import windll
        lang = locale.windows_locale[windll.kernel32.GetUserDefaultUILanguage()]
    if lang is None:
        lang = 'en'
    return lang

def install_gettext():
    lang = get_language()
    log.info('setting language to: %s', lang)
    os.environ['LANGUAGE'] = lang
    locale_dir = pkg_resources.resource_filename('plover', 'gui_qt/messages')
    gettext.install('plover', locale_dir)

def get_gettext(package='plover', resource_dir='gui_qt/messages'):
    locale_dir = pkg_resources.resource_filename(package, resource_dir)
    translation = gettext.translation(package, locale_dir, fallback=True)
    return translation.gettext



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"""Unit tests for formatting.py."""

import inspect

import pytest

from plover import formatting
from plover_build_utils.testing import CaptureOutput

from . import parametrize


def action(**kwargs):
    # Support using something like `text_and_word='stuff'`
    # as a shortcut for `text='stuff', word='stuff'`.
    for k, v in list(kwargs.items()):
        if '_and_' in k:
            del kwargs[k]
            for k in k.split('_and_'):
                kwargs[k] = v
    return formatting._Action(**kwargs)

class MockTranslation:
    def __init__(self, rtfcre=tuple(), english=None, formatting=None):
        self.rtfcre = rtfcre
        self.english = english
        self.formatting = formatting

    def __str__(self):
        return str(self.__dict__)

def translation(**kwargs):
    return MockTranslation(**kwargs)


STARTING_STROKE_TESTS = (

    lambda:
    (True, True, [], [translation(rtfcre=('S'), english='hello')], None,
     ([action(prev_attach=True, text='Hello', trailing_space=' ', word='hello')],),
     [('s', 'Hello')]),

    lambda:
    (False, False, [], [translation(rtfcre=('S'), english='hello')], None,
     ([action(text_and_word='hello', trailing_space=' ')],),
     [('s', ' hello')]),

    lambda:
    (True, False, [], [translation(rtfcre=('S'), english='hello')], None,
     ([action(text='Hello', word='hello', trailing_space=' ')],),
     [('s', ' Hello')]),

    lambda:
    (False, True, [], [translation(rtfcre=('S'), english='hello')], None,
     ([action(text_and_word='hello', prev_attach=True, trailing_space=' ')],),
     [('s', 'hello')]),

)

@parametrize(STARTING_STROKE_TESTS)
def test_starting_stroke(capitalized, attached, undo, do, prev,
                         expected_formats, expected_outputs):
    output = CaptureOutput()
    formatter = formatting.Formatter()
    formatter.set_output(output)
    formatter.start_capitalized = capitalized
    formatter.start_attached = attached
    formatter.format(undo, do, prev)
    for i, t in enumerate(do):
        assert t.formatting == expected_formats[i]
    assert output.instructions == expected_outputs


FORMATTER_TESTS = (

    lambda:
    ([translation(formatting=[action(text_and_word='hello', trailing_space=' ')])],
     [],
     None,
     (),
     [('b', 6)]),

    lambda:
    ([],
     [translation(rtfcre=('S'), english='hello')],
     [translation(rtfcre=('T'), english='a', formatting=[action(text_and_word='f', trailing_space=' ')])]
     ,
     ([action(text_and_word='hello', trailing_space=' ')],),
     [('s', ' hello')]),

    lambda:
    ([], [translation(rtfcre=('S'), english='hello')], None,
     ([action(text_and_word='hello', trailing_space=' ')],),
     [('s', ' hello')]),

    lambda:
    ([], [translation(rtfcre=('ST-T',))], None,
     ([action(text_and_word='ST-T', trailing_space=' ')],),
     [('s', ' ST-T')]),

    lambda:
    ([],
     [translation(rtfcre=('ST-T',))],
     [translation(formatting=[action(text_and_word='hi', trailing_space=' ')])],
     ([action(text_and_word='ST-T', trailing_space=' ')],),
     [('s', ' ST-T')]),

    lambda:
    ([translation(formatting=[action(text_and_word='test', trailing_space=' ')])],
     [translation(english='rest')],
     [translation(formatting=[action(next_case=formatting.CASE_CAP_FIRST_WORD, trailing_space=' ')])],
     ([action(text='Rest', word='rest', trailing_space=' ')],),
     [('b', 4), ('s', 'Rest')]),

    lambda:
    ([translation(formatting=[action(text_and_word='dare'),
                              action(prev_attach=True, text='ing', word='daring', prev_replace='e')])],
     [translation(english='rest')],
     [translation(formatting=[action(next_case=formatting.CASE_CAP_FIRST_WORD,
                                     trailing_space=' ')])],
     ([action(text='Rest', word='rest', trailing_space=' ')],),
     [('b', 6), ('s', 'Rest')]),

    lambda:
    ([translation(formatting=[action(text_and_word='drive', trailing_space=' ')])],
     [translation(english='driving')],
     None,
     ([action(text_and_word='driving', trailing_space=' ')],),
     [('b', 1), ('s', 'ing')]),

    lambda:
    ([translation(formatting=[action(text_and_word='drive', trailing_space=' ')])],
     [translation(english='{#c}driving')],
     None,
     ([action(combo='c'), action(text_and_word='driving', trailing_space=' ')],),
     [('b', 6), ('c', 'c'), ('s', ' driving')]),

    lambda:
    ([translation(formatting=[action(text_and_word='drive', trailing_space=' ')])],
     [translation(english='{PLOVER:c}driving')],
     None,
     ([action(command='c'), action(text_and_word='driving', trailing_space=' ')],),
     [('b', 6), ('e', 'c'), ('s', ' driving')]),

    lambda:
    ([],
     [translation(rtfcre=('1',))],
     None,
     ([action(text_and_word='1', trailing_space=' ', glue=True)],),
     [('s', ' 1')]),

    lambda:
    ([],
     [translation(rtfcre=('1',))],
     [translation(formatting=[action(text_and_word='hi', trailing_space=' ')])],
     ([action(text_and_word='1', trailing_space=' ', glue=True)],),
     [('s', ' 1')]),

    lambda:
    ([],
     [translation(rtfcre=('1',))],
     [translation(formatting=[action(text_and_word='hi', trailing_space=' ', glue=True)])],
     ([action(prev_attach=True, text='1', trailing_space=' ', word='hi1', glue=True)],),
     [('s', '1')]),

    lambda:
    ([],
     [translation(rtfcre=('1-9',))],
     [translation(formatting=[action(text_and_word='hi', trailing_space=' ', glue=True)])],
     ([action(prev_attach=True, text='19', trailing_space=' ', word='hi19', glue=True)],),
     [('s', '19')]),

    lambda:
    ([],
     [translation(rtfcre=('ST-PL',))],
     [translation(formatting=[action(text_and_word='hi', trailing_space=' ')])],
     ([action(text_and_word='ST-PL', trailing_space=' ')],),
     [('s', ' ST-PL')]),

    lambda:
    ([],
     [translation(rtfcre=('ST-PL',))],
     None,
     ([action(text_and_word='ST-PL', trailing_space=' ')],),
     [('s', ' ST-PL')]),

)

@parametrize(FORMATTER_TESTS)
def test_formatter(undo, do, prev, expected_formats, expected_outputs):
    output = CaptureOutput()
    # Add some initial blank text so
    # undoing with no previous state
    # does not assert.
    output.text = ' ' * 128
    formatter = formatting.Formatter()
    formatter.set_output(output)
    formatter.format(undo, do, prev)
    for i, t in enumerate(do):
        assert t.formatting == expected_formats[i]
    assert output.instructions == expected_outputs


def test_action():
    assert action(word='test') != action(word='test', next_attach=True)
    assert action(text='test') == action(text='test')
    assert action(text='test', word='test').copy_state() == action(word='test')


TRANSLATION_TO_ACTIONS_TESTS = (

    lambda:
    ('test', action(),
     [action(text_and_word='test', trailing_space=' ')]),

    lambda:
    ('{^^}', action(),
     [action(prev_attach=True, text_and_word='', next_attach=True, orthography=False)]),

    lambda:
    ('1-9', action(),
     [action(text_and_word='1-9', trailing_space=' ')]),

    lambda:
    ('32', action(),
     [action(text_and_word='32', trailing_space=' ', glue=True)]),

    lambda:
    ('', action(text_and_word='test', next_attach=True),
     [action(prev_attach=True, word='test', next_attach=True)]),

    lambda:
    ('  ', action(text_and_word='test', next_attach=True),
     [action(prev_attach=True, word='test', next_attach=True)]),

    lambda:
    ('{^} {.} hello {.} {#ALT_L(Grave)}{^ ^}', action(),
     [action(prev_attach=True, text_and_word='', next_attach=True, orthography=False),
      action(prev_attach=True, text_and_word='.', trailing_space=' ', next_case=formatting.CASE_CAP_FIRST_WORD),
      action(text='Hello', word='hello', trailing_space=' '),
      action(prev_attach=True, text_and_word='.', trailing_space=' ', next_case=formatting.CASE_CAP_FIRST_WORD),
      action(word='.', trailing_space=' ', combo='ALT_L(Grave)', next_case=formatting.CASE_CAP_FIRST_WORD),
      action(prev_attach=True, text=' ', word='', next_attach=True)
     ]),

    lambda:
    ('{-|}{>}{&a}{>}{&b}', action(),
     [action(next_case=formatting.CASE_CAP_FIRST_WORD),
      action(next_case=formatting.CASE_LOWER_FIRST_CHAR),
      action(text_and_word='a', trailing_space=' ', glue=True),
      action(next_case=formatting.CASE_LOWER_FIRST_CHAR, word='a', trailing_space=' ', glue=True),
      action(prev_attach=True, text='b', word='ab', trailing_space=' ', glue=True),
     ]),

    lambda:
    ('{-|}{>}{&a}{>}{&b}', action(),
     [action(next_case=formatting.CASE_CAP_FIRST_WORD),
      action(next_case=formatting.CASE_LOWER_FIRST_CHAR),
      action(text_and_word='a', trailing_space=' ', glue=True),
      action(next_case=formatting.CASE_LOWER_FIRST_CHAR, word='a', trailing_space=' ', glue=True),
      action(prev_attach=True, text='b', word='ab', trailing_space=' ', glue=True),
     ]),

    lambda:
    ('{-|} equip {^s}', action(),
     [action(next_case=formatting.CASE_CAP_FIRST_WORD),
      action(text='Equip', word='equip', trailing_space=' '),
      action(prev_attach=True, text='s', trailing_space=' ', word='equips'),
     ]),

    lambda:
    ('{-|} equip {^ed}', action(),
     [action(next_case=formatting.CASE_CAP_FIRST_WORD),
      action(text='Equip', word='equip', trailing_space=' '),
      action(prev_attach=True, text='ped', trailing_space=' ', word='equipped'),
     ]),

    lambda:
    ('{>} Equip', action(),
     [action(next_case=formatting.CASE_LOWER_FIRST_CHAR),
      action(text='equip', word='Equip', trailing_space=' ')
     ]),

    lambda:
    ('{>} equip', action(),
     [action(next_case=formatting.CASE_LOWER_FIRST_CHAR),
      action(text_and_word='equip', trailing_space=' ')
     ]),

    lambda:
    ('{<} equip', action(),
     [action(next_case=formatting.CASE_UPPER_FIRST_WORD),
      action(text='EQUIP', word='equip', trailing_space=' ', upper_carry=True)
     ]),

    lambda:
    ('{<} EQUIP', action(),
     [action(next_case=formatting.CASE_UPPER_FIRST_WORD),
      action(text_and_word='EQUIP', trailing_space=' ', upper_carry=True)
     ]),

    lambda:
    ('{<} equip {^ed}', action(),
     [action(next_case=formatting.CASE_UPPER_FIRST_WORD),
      action(text='EQUIP', word='equip', trailing_space=' ', upper_carry=True),
      action(prev_attach=True, text='PED', trailing_space=' ', word='equipped', upper_carry=True)
     ]),

    lambda:
    ('equip {*-|}', action(),
     [action(text_and_word='equip', trailing_space=' '),
      action(prev_attach=True, text='Equip', trailing_space=' ', word_and_prev_replace='equip'),
     ]),

    lambda:
    ('equip {^ed} {*-|}', action(),
     [action(text_and_word='equip', trailing_space=' '),
      action(prev_attach=True, text='ped', trailing_space=' ', word='equipped'),
      action(prev_attach=True, text='Equipped', trailing_space=' ', word_and_prev_replace='equipped'),
     ]),

    lambda:
    ('Equip {*>}', action(),
     [action(text_and_word='Equip', trailing_space=' '),
      action(prev_attach=True, text='equip', trailing_space=' ', word_and_prev_replace='Equip'),
     ]),

    lambda:
    ('Equip {^ed} {*>}', action(),
     [action(text_and_word='Equip', trailing_space=' '),
      action(prev_attach=True, text='ped', trailing_space=' ', word='Equipped'),
      action(prev_attach=True, text='equipped', trailing_space=' ', word_and_prev_replace='Equipped'),
     ]),

    lambda:
    ('equip {*<}', action(),
     [action(text_and_word='equip', trailing_space=' '),
      action(prev_attach=True, text='EQUIP', trailing_space=' ', word_and_prev_replace='equip'),
     ]),

    lambda:
    ('equip {^ed} {*<}', action(),
     [action(text_and_word='equip', trailing_space=' '),
      action(prev_attach=True, text='ped', trailing_space=' ', word='equipped'),
      action(prev_attach=True, text='EQUIPPED', trailing_space=' ', word_and_prev_replace='equipped'),
     ]),

    lambda:
    ('notanumber {*($c)}', action(),
     [action(text_and_word='notanumber', trailing_space=' '),
      action(word='notanumber', trailing_space=' '),
     ]),

    lambda:
    ('0 {*($c)}', action(),
     [action(text_and_word='0', trailing_space=' '),
      action(prev_attach=True, text='$0', word='0', trailing_space=' ', prev_replace='0'),
     ]),

    lambda:
    ('0.00 {*($c)}', action(),
     [action(text_and_word='0.00', trailing_space=' '),
      action(prev_attach=True, text='$0.00', word='0.00', trailing_space=' ', prev_replace='0.00'),
     ]),

    lambda:
    ('1234 {*($c)}', action(),
     [action(text_and_word='1234', trailing_space=' '),
      action(prev_attach=True, text='$1,234', word='1,234', trailing_space=' ', prev_replace='1234'),
     ]),

    lambda:
    ('1234567 {*($c)}', action(),
     [action(text_and_word='1234567', trailing_space=' '),
      action(prev_attach=True, text='$1,234,567', word='1,234,567', trailing_space=' ', prev_replace='1234567'),
     ]),

    lambda:
    ('1234.5 {*($c)}', action(),
     [action(text_and_word='1234.5', trailing_space=' '),
      action(prev_attach=True, text='$1,234.50', word='1,234.50', trailing_space=' ', prev_replace='1234.5'),
     ]),

    lambda:
    ('1234.56 {*($c)}', action(),
     [action(text_and_word='1234.56', trailing_space=' '),
      action(prev_attach=True, text='$1,234.56', word='1,234.56', trailing_space=' ', prev_replace='1234.56'),
     ]),

    lambda:
    ('1234.567 {*($c)}', action(),
     [action(text_and_word='1234.567', trailing_space=' '),
      action(prev_attach=True, text='$1,234.57', word='1,234.57', trailing_space=' ', prev_replace='1234.567'),
     ]),

    lambda:
    ('equip {^} {^ed}', action(),
     [action(text_and_word='equip', trailing_space=' '),
      action(prev_attach=True, text='', word='equip', next_attach=True, orthography=False),
      action(prev_attach=True, text='ed', trailing_space=' ', word='equiped'),
     ]),

    lambda:
     ('{prefix^} test {^ing}', action(),
      [action(text_and_word='prefix', next_attach=True),
       action(prev_attach=True, text_and_word='test', trailing_space=' '),
       action(prev_attach=True, text='ing', trailing_space=' ', word='testing'),
      ]),

    lambda:
    ('{two prefix^} test {^ing}', action(),
     [action(text='two prefix', word='prefix', next_attach=True),
      action(prev_attach=True, text_and_word='test', trailing_space=' '),
      action(prev_attach=True, text='ing', trailing_space=' ', word='testing'),
     ]),

    lambda:
    ('{-|}{^|~|^}', action(),
     [action(next_case=formatting.CASE_CAP_FIRST_WORD),
      action(prev_attach=True, text_and_word='|~|', next_attach=True),
     ]),

    lambda:
    ('{-|}{~|\'^}cause', action(),
     [action(next_case=formatting.CASE_CAP_FIRST_WORD),
      action(text_and_word='\'', next_attach=True, next_case=formatting.CASE_CAP_FIRST_WORD),
      action(prev_attach=True, text='Cause', trailing_space=' ', word='cause'),
     ]),

    lambda:
    ('{.}{~|\'^}cuz', action(),
     [action(prev_attach=True, text_and_word='.', trailing_space=' ', next_case=formatting.CASE_CAP_FIRST_WORD),
      action(text_and_word='\'', next_attach=True, next_case=formatting.CASE_CAP_FIRST_WORD),
      action(prev_attach=True, text='Cuz', trailing_space=' ', word='cuz'),
     ]),

    lambda:
    ('{.}{~|\'^}cause', action(),
     [action(prev_attach=True, text_and_word='.', trailing_space=' ', next_case=formatting.CASE_CAP_FIRST_WORD),
      action(text_and_word='\'', next_attach=True, next_case=formatting.CASE_CAP_FIRST_WORD),
      action(prev_attach=True, text='Cause', trailing_space=' ', word='cause'),
     ]),

    lambda:
    ('{.}{^~|\"}heyyo', action(),
     [action(prev_attach=True, text_and_word='.', trailing_space=' ', next_case=formatting.CASE_CAP_FIRST_WORD),
      action(prev_attach=True, text_and_word='"', trailing_space=' ', next_case=formatting.CASE_CAP_FIRST_WORD),
      action(text='Heyyo', trailing_space=' ', word='heyyo'),
     ]),

    lambda:
    ('{.}{^~|^}zshrc', action(),
     [action(prev_attach=True, text_and_word='.', trailing_space=' ', next_case=formatting.CASE_CAP_FIRST_WORD),
      action(prev_attach=True, text_and_word='', next_attach=True, next_case=formatting.CASE_CAP_FIRST_WORD),
      action(prev_attach=True, text='Zshrc', trailing_space=' ', word='zshrc')]),

    lambda:
    ('{.}', action(),
     [action(prev_attach=True, text_and_word='.', trailing_space=' ', next_case=formatting.CASE_CAP_FIRST_WORD)]
    ),

    lambda:
    ('{,}', action(),
     [action(prev_attach=True, text_and_word=',', trailing_space=' ')]
    ),

    lambda:
    ('test{prefix^}', action(),
     [action(text_and_word='test', trailing_space=' '),
      action(text_and_word='prefix', next_attach=True),
     ]),

    lambda:
    ('{prefix^}{prefix^}', action(),
     [action(text_and_word='prefix', next_attach=True),
      action(prev_attach=True, text='prefix', word='prefixprefix', next_attach=True),
     ]),

    lambda:
    ('{prefix^}{^ing}', action(),
     [action(text_and_word='prefix', next_attach=True),
      action(prev_attach=True, text='ing', trailing_space=' ', word='prefixing'),
     ]),

    lambda:
    ('{prefix^}cancel{^ing}', action(),
     [action(text_and_word='prefix', next_attach=True),
      action(prev_attach=True, text='cancel', trailing_space=' ', word='cancel'),
      action(prev_attach=True, text='ing', trailing_space=' ', word='canceling'),
     ]),

)

@parametrize(TRANSLATION_TO_ACTIONS_TESTS)
def test_translation_to_actions(translation, last_action, expected):
    ctx = formatting._Context([], action())
    ctx.translated(last_action)
    assert formatting._translation_to_actions(translation, ctx) == expected


RAW_TO_ACTIONS_TESTS = (

    lambda:
    ('2-6', action(),
     [action(glue=True, text_and_word='26', trailing_space=' ')]),

    lambda:
    ('2', action(),
     [action(glue=True, text_and_word='2', trailing_space=' ')]),

    lambda:
    ('-8', action(),
     [action(glue=True, text_and_word='8', trailing_space=' ')]),

    lambda:
    ('-68', action(),
     [action(glue=True, text_and_word='68', trailing_space=' ')]),

    lambda:
    ('S-T', action(),
     [action(text_and_word='S-T', trailing_space=' ')]),

)

@parametrize(RAW_TO_ACTIONS_TESTS)
def test_raw_to_actions(stroke, last_action, expected):
    ctx = formatting._Context([], action())
    ctx.translated(last_action)
    assert formatting._raw_to_actions(stroke, ctx) == expected


ATOM_TO_ACTION_TESTS = (

    lambda:
    ('{^ed}', action(text_and_word='test', trailing_space=' '),
     action(prev_attach=True, text='ed', trailing_space=' ', word='tested')),

    lambda:
    ('{^ed}', action(text_and_word='carry', trailing_space=' '),
     action(prev_attach=True, text='ied', trailing_space=' ', prev_replace='y', word='carried')),

    lambda:
    ('{^er}', action(text_and_word='test', trailing_space=' '),
     action(prev_attach=True, text='er', trailing_space=' ', word='tester')),

    lambda:
    ('{^er}', action(text_and_word='carry', trailing_space=' '),
     action(prev_attach=True, text='ier', trailing_space=' ', prev_replace='y', word='carrier')),

    lambda:
    ('{^ing}', action(text_and_word='test', trailing_space=' '),
     action(prev_attach=True, text='ing', trailing_space=' ', word='testing')),

    lambda:
    ('{^ing}', action(text_and_word='begin', trailing_space=' '),
     action(prev_attach=True, text='ning', trailing_space=' ', word='beginning')),

    lambda:
    ('{^ing}', action(text_and_word='parade', trailing_space=' '),
     action(prev_attach=True, text='ing', trailing_space=' ', prev_replace='e', word='parading')),

    lambda:
    ('{^s}', action(text_and_word='test', trailing_space=' '),
     action(prev_attach=True, text='s', trailing_space=' ', word='tests')),

    lambda:
    ('{,}', action(text_and_word='test', trailing_space=' '),
     action(prev_attach=True, text_and_word=',', trailing_space=' ')),

    lambda:
    ('{:}', action(text_and_word='test', trailing_space=' '),
     action(prev_attach=True, text_and_word=':', trailing_space=' ')),

    lambda:
    ('{;}', action(text_and_word='test', trailing_space=' '),
     action(prev_attach=True, text_and_word=';', trailing_space=' ')),

    lambda:
    ('{.}', action(prev_attach=True, word='test', trailing_space=' '),
     action(prev_attach=True, text_and_word='.', trailing_space=' ', next_case=formatting.CASE_CAP_FIRST_WORD)),

    lambda:
    ('{?}', action(text_and_word='test', trailing_space=' '),
     action(prev_attach=True, text_and_word='?', trailing_space=' ', next_case=formatting.CASE_CAP_FIRST_WORD)),

    lambda:
    ('{!}', action(text_and_word='test', trailing_space=' '),
     action(prev_attach=True, text_and_word='!', trailing_space=' ', next_case=formatting.CASE_CAP_FIRST_WORD)),

    lambda:
    ('{-|}', action(text_and_word='test', trailing_space=' '),
     action(next_case=formatting.CASE_CAP_FIRST_WORD, word='test', trailing_space=' ')),

    lambda:
    ('{>}', action(text_and_word='test', trailing_space=' '),
     action(next_case=formatting.CASE_LOWER_FIRST_CHAR, word='test', trailing_space=' ')),

    lambda:
    ('{<}', action(text_and_word='test', trailing_space=' '),
     action(next_case=formatting.CASE_UPPER_FIRST_WORD, word='test', trailing_space=' ')),

    lambda:
    ('{*-|}', action(text_and_word='test', trailing_space=' '),
     action(prev_attach=True, text='Test', trailing_space=' ', word_and_prev_replace='test')),

    lambda:
    ('{*>}', action(text_and_word='test', trailing_space=' '),
     action(prev_attach=True, text_and_word_and_prev_replace='test',
            trailing_space=' ')),

    lambda:
    ('{*<}', action(text_and_word='test', trailing_space=' '),
     action(prev_attach=True, text='TEST', word_and_prev_replace='test', trailing_space=' ')),

    lambda:
    ('{PLOVER:test_command}', action(text_and_word='test', trailing_space=' '),
     action(word='test', command='test_command', trailing_space=' ')),

    lambda:
    ('{&glue_text}', action(text_and_word='test', trailing_space=' '),
     action(text_and_word='glue_text', trailing_space=' ', glue=True)),

    lambda:
    ('{&glue_text}', action(text_and_word='test', trailing_space=' ', glue=True),
     action(prev_attach=True, text='glue_text', trailing_space=' ', word='testglue_text', glue=True)),

    lambda:
    ('{&glue_text}', action(text_and_word='test', next_attach=True),
     action(prev_attach=True, text='glue_text', trailing_space=' ', word='glue_text', glue=True)),

    lambda:
    ('{^attach_text}', action(text_and_word='test', trailing_space=' '),
     action(prev_attach=True, text='attach_text', trailing_space=' ', word='testattach_text')),

    lambda:
    ('{^attach_text^}', action(text_and_word='test', trailing_space=' '),
     action(prev_attach=True, text='attach_text', word='testattach_text', next_attach=True)),

    lambda:
    ('{attach_text^}', action(text_and_word='test', trailing_space=' '),
     action(text_and_word='attach_text', next_attach=True)),

    lambda:
    ('{#ALT_L(A)}', action(text_and_word='test', trailing_space=' '),
     action(combo='ALT_L(A)', trailing_space=' ', word='test')),

    lambda:
    ('text', action(text_and_word='test', trailing_space=' '),
     action(text_and_word='text', trailing_space=' ')),

    lambda:
    ('text', action(text_and_word='test', trailing_space=' ', glue=True),
     action(text_and_word='text', trailing_space=' ')),

    lambda:
    ('text', action(text_and_word='test', next_attach=True),
     action(prev_attach=True, text_and_word='text', trailing_space=' ')),

    lambda:
    ('text', action(text_and_word='test', trailing_space=' ', next_case=formatting.CASE_CAP_FIRST_WORD),
     action(text='Text', trailing_space=' ', word='text')),

    lambda:
    ('some text', action(text_and_word='test', trailing_space=' '),
     action(text='some text', trailing_space=' ', word='text')),

    lambda:
    ('some text',
     action(text_and_word='test', trailing_space=' ',
            case=formatting.CASE_TITLE,
            space_char=''),
     action(text='SomeText', word='text',
            case=formatting.CASE_TITLE,
            space_char='')),

    lambda:
    ('some text',
     action(text_and_word='test', trailing_space=' ', # This is camel case
            case=formatting.CASE_TITLE,
            space_char='', next_case=formatting.CASE_LOWER_FIRST_CHAR),
     action(text='someText', word='text',
            case=formatting.CASE_TITLE,
            space_char='')),

    lambda:
    ('some text', action(text_and_word='test', trailing_space=' ', space_char='_'),
     action(text='some_text', trailing_space='_', word='text', space_char='_')),

    lambda:
    ('some text', action(text_and_word='test', trailing_space=' ', case=formatting.CASE_UPPER),
     action(text='SOME TEXT', trailing_space=' ', word='text', case=formatting.CASE_UPPER)),

    lambda:
    ('sOme TexT', action(text_and_word='test', trailing_space=' ', case=formatting.CASE_LOWER),
     action(text='some text', trailing_space=' ', word='TexT', case=formatting.CASE_LOWER)),

    lambda:
    ('sOme TexT', action(text_and_word='test', trailing_space=' ', case=formatting.CASE_TITLE),
     action(text='Some Text', trailing_space=' ', word='TexT', case=formatting.CASE_TITLE)),

    lambda:
    ('{MODE:CAPS}', action(text_and_word='test', trailing_space=' '),
     action(word='test', trailing_space=' ', case=formatting.CASE_UPPER)),

    lambda:
    ('{MODE:LOWER}', action(text_and_word='test', trailing_space=' '),
     action(word='test', trailing_space=' ', case=formatting.CASE_LOWER)),

)

@parametrize(ATOM_TO_ACTION_TESTS)
def test_atom_to_action(atom, last_action, expected):
    ctx = formatting._Context((), last_action)
    ctx.translated(last_action)
    assert formatting._atom_to_action(atom, ctx) == expected


CHANGE_MODE_TESTS = (
    # Undefined
    lambda:
    ('', action(),
     action()),
    lambda:
    ('ABCD', action(),
     action()),
    # CAPS: Uppercase
    lambda:
    ('CAPS', action(),
     action(case=formatting.CASE_UPPER)),
    # LOWER: Lowercase
    lambda:
    ('LOWER', action(),
     action(case=formatting.CASE_LOWER)),
    # TITLE: Titlecase
    lambda:
    ('TITLE', action(),
     action(case=formatting.CASE_TITLE)),
    # CAMEL: Titlecase without space
    lambda:
    ('CAMEL', action(),
     action(case=formatting.CASE_TITLE, space_char='',
            next_case=formatting.CASE_LOWER_FIRST_CHAR)),
    # SNAKE: Underscore space
    lambda:
    ('SNAKE', action(),
     action(space_char='_')),
    # RESET_SPACE: Default space
    lambda:
    ('RESET_SPACE', action(space_char='ABCD'),
     action()),
    # RESET_CASE: No case
    lambda:
    ('RESET_CASE', action(case=formatting.CASE_UPPER),
     action()),
    # SET_SPACE:xy: Set space to xy
    lambda:
    ('SET_SPACE:', action(space_char='test'),
     action(space_char='')),
    lambda:
    ('SET_SPACE:-', action(space_char='test'),
     action(space_char='-')),
    lambda:
    ('SET_SPACE:123 45', action(space_char='test'),
     action(space_char='123 45')),
    # RESET: No case, default space
)

@parametrize(CHANGE_MODE_TESTS)
def test_meta_mode(meta, last_action, expected):
    ctx = formatting._Context((), action())
    ctx.translated(last_action)
    assert formatting._atom_to_action('{MODE:' + meta + '}', ctx) == expected


last_action_normal = action()
last_action_capitalized = action(next_case=formatting.CASE_CAP_FIRST_WORD)
last_action_attached = action(next_attach=True)

META_CARRY_CAPITALIZE_TESTS = (

    # Test word handling and space handling, standard.
    lambda:
    ('~|*', last_action_normal,
     (action(word='*', text='*', trailing_space=' '))),

    # With attach flags:
    lambda:
    ('~|*^', last_action_normal,
     (action(word='*', text='*', next_attach=True))),
    lambda:
    ('^~|*', last_action_normal,
     (action(word='*', text='*', trailing_space=' ', prev_attach=True))),
    lambda:
    ('^~|*^', last_action_normal,
     (action(word='*', text='*', prev_attach=True, next_attach=True))),

    # Should 'do nothing'.
    lambda:
    ('~|', last_action_capitalized,
     (last_action_capitalized)),

    # Should lose 'next_attach' flag.
    lambda:
    ('~|', last_action_attached,
     (action(prev_attach=True))),

    # Verify capitalize carry.
    lambda:
    ('^~|^', last_action_capitalized,
     (action(next_case=formatting.CASE_CAP_FIRST_WORD, text_and_word='', prev_attach=True, next_attach=True))),
    lambda:
    ('^~|aset^', last_action_capitalized,
     (action(next_case=formatting.CASE_CAP_FIRST_WORD, prev_attach=True, next_attach=True, text='Aset', word='aset'))),
    lambda:
    ('~|aset', last_action_capitalized,
     (action(next_case=formatting.CASE_CAP_FIRST_WORD, text='Aset', trailing_space=' ', word='aset'))),

    # Verify 'next_attach' flag overriding.
    lambda:
    ('~|aset', last_action_attached,
     (action(prev_attach=True, text_and_word='aset', trailing_space=' '))),
    lambda:
    ('~|aset^', last_action_attached,
     (action(prev_attach=True, text_and_word='aset', next_attach=True))),

)

@parametrize(META_CARRY_CAPITALIZE_TESTS)
def test_meta_carry_capitalize(meta, last_action, expected):
    ctx = formatting._Context((), action())
    ctx.translated(last_action)
    assert formatting._atom_to_action('{' + meta + '}', ctx) == expected


def _apply_case_tests():
    test = ' some test '
    test2 = 'test Me'
    test3 = ' SOME TEST '
    return (
        # INVALID
        lambda: (test, '', False, ValueError),
        lambda: (test, 'TEST', False, ValueError),
        # NO-OP
        lambda: (test, None, False, test),
        lambda: (test, None, True, test),
        # TITLE
        lambda: (test, formatting.CASE_TITLE, False, ' Some Test '),
        # TITLE will not affect appended output
        lambda: (test, formatting.CASE_TITLE, True, ' some test '),
        lambda: (test2, formatting.CASE_TITLE, True, 'test Me'),
        # LOWER
        lambda: (test, formatting.CASE_LOWER, False, ' some test '),
        lambda: (test3, formatting.CASE_LOWER, False, ' some test '),
        lambda: (test2, formatting.CASE_LOWER, True, 'test me'),
        # UPPER
        lambda: (test.upper(), formatting.CASE_UPPER, False, ' SOME TEST '),
        lambda: (test3, formatting.CASE_UPPER, False, ' SOME TEST '),
        lambda: (test2, formatting.CASE_UPPER, True, 'TEST ME'),
    )

@parametrize(_apply_case_tests())
def test_apply_case(input_text, case, appended, expected):
    if inspect.isclass(expected):
        with pytest.raises(expected):
            formatting._apply_mode_case(input_text, case, appended)
    else:
        assert formatting._apply_mode_case(input_text, case, appended) == expected


def _apply_space_char_tests():
    test = ' some text '
    test2 = "don't"
    return (
        lambda: (test, '_', '_some_text_'),
        lambda: (test, '', 'sometext'),
        lambda: (test2, '_', test2),
        lambda: (test2, '', test2),
    )

@parametrize(_apply_space_char_tests())
def test_apply_space_char(text, space_char, expected):
    assert formatting._apply_mode_space_char(text, space_char) == expected


@parametrize((
    lambda: ('', None),
    lambda: ('{abc}', 'abc'),
    lambda: ('abc', None),
))
def test_get_meta(atom, meta):
    assert formatting._get_meta(atom) == meta


@parametrize((
    lambda: ('abc', '{&abc}'),
    lambda: ('1', '{&1}'),
))
def test_glue_translation(s, expected):
    assert formatting._glue_translation(s) == expected


@parametrize((
    lambda: ('', ''),
    lambda: ('abc', 'abc'),
    lambda: (r'\{', '{'),
    lambda: (r'\}', '}'),
    lambda: (r'\{abc\}}{', '{abc}}{'),
))
def test_unescape_atom(atom, text):
    assert formatting._unescape_atom(atom) == text


@parametrize((
    lambda: ('', ''),
    lambda: ('abc', 'Abc'),
    lambda: ('ABC', 'ABC'),
))
def test_capitalize_first_word(s, expected):
    assert formatting._capitalize_first_word(s) == expected


RIGHTMOST_WORD_TESTS = (
    lambda: ('', ''),
    lambda: ('\n', ''),
    lambda: ('\t', ''),
    lambda: ('abc', 'abc'),
    lambda: ('a word', 'word'),
    lambda: ('word.', '.'),
    lambda: ('word ', ''),
    lambda: ('word\n', ''),
    lambda: ('word\t', ''),
    lambda: (' word', 'word'),
    lambda: ('\nword', 'word'),
    lambda: ('\tword', 'word'),
)

@parametrize(RIGHTMOST_WORD_TESTS)
def test_rightmost_word(s, expected):
    assert formatting._rightmost_word(s) == expected


REPLACE_TESTS = (

    # Check that 'prev_replace' does not unconditionally erase
    # the previous character if it does not match.
    lambda:
    ([
        translation(english='{MODE:SET_SPACE:}'),
        translation(english='foobar'),
        translation(english='{^}{#Return}{^}{-|}'),

    ], [('s', 'foobar'), ('c', 'Return')]),

    # Check 'prev_replace' correctly takes into account
    # the previous translation.
    lambda:
    ([
        translation(english='test '),
        translation(english='{^,}'),

    ], [('s', 'test '), ('b', 1), ('s', ', ')]),

    # While the previous translation must be taken into account,
    # any meta-command must not be fired again.
    lambda:
    ([
        translation(english='{#Return}'),
        translation(english='test'),

    ], [('c', 'Return'), ('s', 'test ')]),

)

@parametrize(REPLACE_TESTS)
def test_replace(translations, expected_instructions):
    output = CaptureOutput()
    formatter = formatting.Formatter()
    formatter.set_output(output)
    formatter.set_space_placement('After Output')
    prev = []
    for t in translations:
        formatter.format([], [t], prev)
        prev.append(t)
    assert output.instructions == expected_instructions


def test_undo_replace():
    # Undoing a replace....
    output = CaptureOutput()
    formatter = formatting.Formatter()
    formatter.set_output(output)
    formatter.set_space_placement('After Output')
    prev = [translation(english='test')]
    formatter.format([], prev, None)
    undo = translation(english='{^,}')
    formatter.format([], [undo], prev)
    # Undo.
    formatter.format([undo], [], prev)
    assert output.instructions == [
        ('s', 'test '), ('b', 1), ('s', ', '), ('b', 2), ('s', ' '),
    ]


OUTPUT_OPTIMISATION_TESTS = (

    # No change.
    lambda:
    ([
        translation(english='noop'),
    ], [
        translation(english='noop'),
    ], [('s', ' noop')]),

    # Append only.
    lambda:
    ([
        translation(english='test'),
    ], [
        translation(english='testing'),
    ], [('s', ' test'), ('s', 'ing')]),

    # Chained meta-commands.
    lambda:
    ([
        translation(english='{#a}'),
    ], [
        translation(english='{#a}{#b}'),
    ], [('c', 'a'), ('c', 'b')]),

)

@parametrize(OUTPUT_OPTIMISATION_TESTS)
def test_output_optimization(undo, do, expected_instructions):
    output = CaptureOutput()
    formatter = formatting.Formatter()
    formatter.set_output(output)
    formatter.format([], undo, None)
    formatter.format(undo, do, None)
    assert output.instructions == expected_instructions


class TestRetroFormatter:

    def setup_method(self):
        self.formatter = formatting.Formatter()
        self.translations = []
        self.retro_formatter = formatting.RetroFormatter(self.translations)

    def format(self, text):
        t = translation(english=text)
        self.formatter.format([], [t], self.translations)
        self.translations.append(t)
        return t

    ITER_LAST_ACTIONS_TESTS = (

        lambda:
        (['Luca', 'mela'],
         [action(text_and_word='mela', trailing_space=' '),
          action(text_and_word='Luca', trailing_space=' ')]),

        lambda:
        (['{Luca^}', '{^mela}'],
         [action(prev_attach=True, text='mela', trailing_space=' ', word='Lucamela'),
          action(text_and_word='Luca', next_attach=True)]),

        lambda:
        (['Luca', '{^ ^}', 'mela'],
         [action(text_and_word='mela', trailing_space=' ', prev_attach=True),
          action(text=' ', word='', prev_attach=True, next_attach=True),
          action(text_and_word='Luca', trailing_space=' ')]),

        lambda:
        (['Luca', '{-|}', 'mela'],
         [action(text='Mela', trailing_space=' ', word='mela'),
          action(word='Luca', trailing_space=' ', next_case=formatting.CASE_CAP_FIRST_WORD),
          action(text_and_word='Luca', trailing_space=' ')]),

    )

    @parametrize(ITER_LAST_ACTIONS_TESTS)
    def test_iter_last_actions(self, translation_list, action_list):
        for t in translation_list:
            self.format(t)
        assert list(self.retro_formatter.iter_last_actions()) == action_list

    ITER_LAST_FRAGMENTS_TESTS = (

        lambda:
        (False,
         ['Luca', 'mela'],
         ['mela', 'Luca ']),

        lambda:
        (False,
         ['Luca{^ ^}mela'],
         ['mela', 'Luca ']),

        lambda:
        (False,
         ['Luca, mela.'],
         ['mela.', 'Luca, ']),

        lambda:
        (False,
         ['Luca{-|}mela'],
         ['Mela', 'Luca ']),

        lambda:
        (True,
         ['Luca{-|}mela'],
         ['Mela', 'Luca ']),

    )

    @parametrize(ITER_LAST_FRAGMENTS_TESTS)
    def test_iter_last_fragments(self, spaces_after, translation_list, fragment_list):
        if spaces_after:
            self.formatter.set_space_placement('After Output')
        for t in translation_list:
            self.format(t)
        assert list(self.retro_formatter.iter_last_fragments()) == fragment_list

    ITER_LAST_WORDS_TESTS = (

        lambda:
        (False,
         ['Luca', 'mela'],
         ['mela', 'Luca ']),

        lambda:
        (False,
         ['Luca{^ ^}mela'],
         ['mela', 'Luca ']),

        lambda:
        (False,
         ['Luca, mela.'],
         ['.', 'mela', ', ', 'Luca']),

        lambda:
        (False,
         ['Luca{-|}mela'],
         ['Mela', 'Luca ']),

        lambda:
        (True,
         ['Luca{-|}mela'],
         ['Mela', 'Luca ']),
    )

    @parametrize(ITER_LAST_WORDS_TESTS)
    def test_iter_last_words(self, spaces_after, translation_list, word_list):
        if spaces_after:
            self.formatter.set_space_placement('After Output')
        for t in translation_list:
            self.format(t)
        assert list(self.retro_formatter.iter_last_words()) == word_list

    LAST_TEXT_TESTS = (

        lambda:
        (False,
         ['Luca{-|}mela'],
         3,
         'ela'),

        lambda:
        (False,
         ['Luca{-|}mela'],
         5,
         ' Mela'),

        lambda:
        (False,
         ['Luca{-|}mela'],
         12,
         'Luca Mela'),

        lambda:
        (False,
         ['Luca{-|}mela'],
         20,
         'Luca Mela'),

        lambda:
        (True,
         ['Luca{-|}mela'],
         6,
         'a Mela'),

    )

    @parametrize(LAST_TEXT_TESTS)
    def test_last_text(self, spaces_after, translation_list, count, text):
        if spaces_after:
            self.formatter.set_space_placement('After Output')
        for t in translation_list:
            self.format(t)
        assert self.retro_formatter.last_text(count) == text



#!/usr/bin/env python3
# Copyright (c) 2010 Joshua Harlan Lifton.
# See LICENSE.txt for details.
#
# keyboardcontrol.py - Abstracted keyboard control.
#
# Uses OS appropriate module.

"""Keyboard capture and control.

This module provides an interface for basic keyboard event capture and
emulation. Set the key_up and key_down functions of the
KeyboardCapture class to capture keyboard input. Call the send_string
and send_backspaces functions of the KeyboardEmulation class to
emulate keyboard input.

"""

import sys

KEYBOARDCONTROL_NOT_FOUND_FOR_OS = \
        "No keyboard control module was found for os %s" % sys.platform

if sys.platform.startswith('linux'):
    from plover.oslayer import xkeyboardcontrol as keyboardcontrol
elif sys.platform.startswith('win32'):
    from plover.oslayer import winkeyboardcontrol as keyboardcontrol
elif sys.platform.startswith('darwin'):
    from plover.oslayer import osxkeyboardcontrol as keyboardcontrol
else:
    raise Exception(KEYBOARDCONTROL_NOT_FOUND_FOR_OS)


class KeyboardCapture(keyboardcontrol.KeyboardCapture):
    """Listen to keyboard events."""

    # Supported keys.
    SUPPORTED_KEYS_LAYOUT = '''
    Escape  F1 F2 F3 F4  F5 F6 F7 F8  F9 F10 F11 F12

      `  1  2  3  4  5  6  7  8  9  0  -  =  \\ BackSpace  Insert Home Page_Up
     Tab  q  w  e  r  t  y  u  i  o  p  [  ]               Delete End  Page_Down
           a  s  d  f  g  h  j  k  l  ;  '      Return
            z  x  c  v  b  n  m  ,  .  /                          Up
                     space                                   Left Down Right
    '''
    SUPPORTED_KEYS = tuple(SUPPORTED_KEYS_LAYOUT.split())


class KeyboardEmulation(keyboardcontrol.KeyboardEmulation):
    """Emulate printable key presses and backspaces."""
    pass


if __name__ == '__main__':

    import time

    kc = KeyboardCapture()
    ke = KeyboardEmulation()

    pressed = set()
    status = 'pressed: '

    def test(key, action):
        global status
        print(key, action)
        if 'pressed' == action:
            pressed.add(key)
        elif key in pressed:
            pressed.remove(key)
        new_status = 'pressed: ' + '+'.join(pressed)
        if status != new_status:
            ke.send_backspaces(len(status))
            ke.send_string(new_status)
            status = new_status

    kc.key_down = lambda k: test(k, 'pressed')
    kc.key_up = lambda k: test(k, 'released')
    kc.suppress_keyboard(KeyboardCapture.SUPPORTED_KEYS)
    kc.start()
    print('Press CTRL-c to quit.')
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        kc.cancel()




from PyQt5.QtCore import QSettings
from PyQt5.QtWidgets import (
    QMainWindow,
    QToolBar,
    QToolButton,
    QWidget,
)


def ToolButton(action):
    button = QToolButton()
    button.setDefaultAction(action)
    return button


def ToolBar(*action_list):
    toolbar = QToolBar()
    for action in action_list:
        if action is None:
            toolbar.addSeparator()
        else:
            toolbar.addWidget(ToolButton(action))
    return toolbar


class WindowState(QWidget):

    ROLE = None

    def _save_state(self, settings):
        pass

    def save_state(self):
        assert self.ROLE
        settings = QSettings()
        settings.beginGroup(self.ROLE)
        settings.setValue('geometry', self.saveGeometry())
        if isinstance(self, QMainWindow):
            settings.setValue('state', self.saveState())
        self._save_state(settings)
        settings.endGroup()

    def _restore_state(self, settings):
        pass

    def restore_state(self):
        assert self.ROLE
        settings = QSettings()
        settings.beginGroup(self.ROLE)
        geometry = settings.value('geometry')
        if geometry is not None:
            self.restoreGeometry(geometry)
        if isinstance(self, QMainWindow):
            state = settings.value('state')
            if state is not None:
                self.restoreState(state)
        self._restore_state(settings)
        settings.endGroup()


def find_menu_actions(menu):
    '''Recursively find and return a menu actions.'''
    actions_dict = {}
    for action in menu.actions():
        name = action.objectName()
        if not name:
            sub_menu = action.menu()
            if sub_menu is None:
                continue
            actions_dict.update(find_menu_actions(sub_menu))
            name = sub_menu.objectName()
        assert name
        assert name not in actions_dict
        actions_dict[name] = action
    return actions_dict




import inspect
import itertools

import pytest

from plover.key_combo import parse_key_combo


def generate_combo_tests(test_id, *params, key_name_to_key_code=None):
    yield ('id', test_id)
    if key_name_to_key_code is not None:
        yield ('key_name_to_key_code', key_name_to_key_code)
    for iterables in params:
        iterables = [
            i if isinstance(i, (tuple, list)) else (i,)
            for i in iterables
        ]
        if len(iterables) < 2:
            iterables.append(('',))
        for combo_string, parse_result, in itertools.product(*iterables):
            yield (
                ('parse', combo_string, parse_result),
            )


# Test directives:
# - id TEST_IDENTIFIER
# - key_name_to_key_code KEY_NAME_TO_KEY_CODE_FUNCTION
# - parse COMBO_STRING KEY_EVENTS
KEY_COMBO_TESTS = tuple(itertools.chain(

    (
        ('key_name_to_key_code', lambda k: k),
    ),

    # No-op.
    generate_combo_tests(
        'no-op',
        (('', '   '), ''),
    ),

    # Syntax error:
    generate_combo_tests(
        'syntax_error',
        ((
            # - invalid character
            'Return,',
            'Return&',
            'Ret. urn <',
            'exclam ! foo',
            'shift[a]',
            # - unbalanced `)`
            ') arg',
            'arg )',
            'arg())',
            'arg(x) )',
            # - unbalanced `(`
            'test(',
            '( grr',
            'foo ( bar',
            'foo (bar ( ',
            'foo ((',
            # - [-+]key() is not valid
            '+foo ()',
            '+foo()',
        ), SyntaxError),
    ),

    # Pressing an already pressed key.
    generate_combo_tests(
        'already_pressed',
        ((
            'foo(foo)',
            'Foo(foO)',
            'foo(fOo(arg))',
            'foo(bar(Foo))',
            'foo(bar(foo(x)))',
        ), ValueError),
    ),

    # Stacking.
    generate_combo_tests(
        'stacking',
        # 1 is not a valid identifier, but still a valid key name.
        ('1', '+1 -1'),
        (('Shift_l', 'SHIFT_L'), '+shift_l -shift_l'),
        # Case does not matter.
        (('a', ' A '), '+a -a'),
        (('a(b c)', 'a ( b c   )'), '+a +b -b +c -c -a'),
        (('a(bc)', ' a(  Bc )'), '+a +bc -bc -a'),
        (('a(bc(d)e f(g) h())i j'),
         '+a +bc +d -d -bc +e -e +f +g -g -f +h -h -a +i -i +j -j'),
        (('foo () bar ( foo a b c (d))', 'fOo () Bar ( FOO a B c (D))'),
         '+foo -foo +bar +foo -foo +a -a +b -b +c +d -d -c -bar'),
    ),

    # Invalid key name.
    generate_combo_tests(
        'invalid_key',
        ('1 (c) 2 bad 3 (a b c)', ValueError),
        key_name_to_key_code={c: c for c in '123abc'}.get,
    ),

    # Same key code, multiple key names.
    generate_combo_tests(
        'aliasing',
        ('1 exclam', '+10 -10 +10 -10'),
        (('1 ( exclam )', 'exclam(1)'), ValueError),
        key_name_to_key_code={'1': '10', 'exclam': '10'}.get,
    ),

))

def parametrize(tests):
    key_name_to_key_code = None
    test_id = None
    args = []
    ids = []
    for t in tests:
        if t[0] == 'key_name_to_key_code':
            key_name_to_key_code = t[1]
        elif t[0] == 'id':
            test_id = t[1]
        else:
            assert key_name_to_key_code is not None
            assert test_id is not None
            args.append((key_name_to_key_code, t))
            ids.append(test_id)
    return pytest.mark.parametrize(
        ('key_name_to_key_code', 'instructions'),
        args, ids=ids
    )

@parametrize(KEY_COMBO_TESTS)
def test_key_combo(key_name_to_key_code, instructions):
    def repr_expected(result):
        assert isinstance(result, str)
        return [s.strip() for s in result.split()]
    def repr_key_events(events):
        assert isinstance(events, list)
        return ['%s%s' % ('+' if pressed else '-', key)
                for key, pressed in events]
    for action, *args in instructions:
        if action == 'parse':
            combo_string, key_events = args
            if inspect.isclass(key_events):
                with pytest.raises(key_events):
                    parse_key_combo(combo_string, key_name_to_key_code=key_name_to_key_code)
            else:
                assert repr_key_events(parse_key_combo(combo_string, key_name_to_key_code=key_name_to_key_code)) == repr_expected(key_events)
        else:
            raise ValueError(args[0])



import re


def fix_icons(contents):
    # replace ``addPixmap(QtGui.QPixmap(":/settings.svg"),``
    # by ``addFile(":/settings.svg", QtCore.QSize(),``
    contents = re.sub(r'\baddPixmap\(QtGui.QPixmap\(("[^"]*")\),',
                      r'addFile(\1, QtCore.QSize(),', contents)
    return contents

def gettext(contents):
    # replace ``_translate("context", `` by ``_(``
    contents = re.sub(r'_translate\(".*",\s', '_(', contents)
    contents = contents.replace(
        '        _translate = QtCore.QCoreApplication.translate',
        ''
    )
    return contents



from PyQt5.QtGui import (
    QFont,
    QTextCursor,
    QTextCharFormat,
)
from PyQt5.QtWidgets import QWidget

from plover.translation import escape_translation

from plover.gui_qt.suggestions_widget_ui import Ui_SuggestionsWidget


class SuggestionsWidget(QWidget, Ui_SuggestionsWidget):

    STYLE_TRANSLATION, STYLE_STROKES = range(2)

    # Anatomy of the text document:
    # - "root":
    #  - 0+ "suggestions" blocks
    #   - 1+ "translation" blocks
    #    - 1-10 "strokes" blocks

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setupUi(self)
        self._translation_char_format = QTextCharFormat()
        self._strokes_char_format = QTextCharFormat()
        self._strokes_char_format.font().setStyleHint(QFont.Monospace)

    def append(self, suggestion_list):
        scrollbar = self.suggestions.verticalScrollBar()
        scroll_at_end = scrollbar.value() == scrollbar.maximum()
        cursor = self.suggestions.textCursor()
        cursor.movePosition(QTextCursor.End)
        for suggestion in suggestion_list:
            cursor.insertBlock()
            cursor.setCharFormat(self._translation_char_format)
            cursor.block().setUserState(self.STYLE_TRANSLATION)
            cursor.insertText(escape_translation(suggestion.text) + ':')
            if not suggestion.steno_list:
                cursor.insertText(' ' + _('no suggestions'))
                continue
            for strokes_list in suggestion.steno_list[:10]:
                cursor.insertBlock()
                cursor.setCharFormat(self._strokes_char_format)
                cursor.block().setUserState(self.STYLE_STROKES)
                cursor.insertText('   ' + '/'.join(strokes_list))
        cursor.insertText('\n')
        # Keep current position when not at the end of the document.
        if scroll_at_end:
            scrollbar.setValue(scrollbar.maximum())

    def clear(self):
        self.suggestions.clear()

    def _reformat(self):
        document = self.suggestions.document()
        cursor = self.suggestions.textCursor()
        block = document.begin()
        style_format = {
            self.STYLE_TRANSLATION: self._translation_char_format,
            self.STYLE_STROKES: self._strokes_char_format,
        }
        while block != document.end():
            style = block.userState()
            fmt = style_format.get(style)
            if fmt is not None:
                cursor.setPosition(block.position())
                cursor.select(QTextCursor.BlockUnderCursor)
                cursor.setCharFormat(fmt)
            block = block.next()

    @property
    def text_font(self):
        return self._translation_char_format.font()

    @text_font.setter
    def text_font(self, font):
        self._translation_char_format.setFont(font)
        self._reformat()

    @property
    def strokes_font(self):
        return self._strokes_char_format.font()

    @strokes_font.setter
    def strokes_font(self, font):
        self._strokes_char_format.setFont(font)
        self._reformat()



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

import os
from logging import Handler
from collections import defaultdict

import pytest

from plover.steno import Stroke
from plover import log


class FakeHandler(Handler):

    outputs = defaultdict(list)

    def __init__(self, filename, format=log.STROKE_LOG_FORMAT):
        super().__init__()
        self.baseFilename = filename

    def emit(self, record):
        FakeHandler.outputs[self.baseFilename].append(record.getMessage())


@pytest.fixture(autouse=True)
def fake_file_log(monkeypatch):
    monkeypatch.setattr('plover.log.FileHandler', FakeHandler)
    yield
    FakeHandler.outputs.clear()
    # Reset logger state.
    log.set_stroke_filename(None)
    log.enable_stroke_logging(False)
    log.enable_translation_logging(False)


def stroke_filename(path):
    return os.path.abspath(path)


def test_set_filename():
    sf1 = stroke_filename('/fn1')
    log.set_stroke_filename('/fn1')
    log.enable_stroke_logging(True)
    log.stroke(Stroke(('S-',)))
    sf2 = stroke_filename('/fn2')
    log.set_stroke_filename('/fn2')
    log.stroke(Stroke(('-T',)))
    log.set_stroke_filename(None)
    log.stroke(Stroke(('P-',)))
    assert FakeHandler.outputs == {
        sf1: ["Stroke(S : ['S-'])"],
        sf2: ["Stroke(-T : ['-T'])"],
    }

def test_stroke():
    sf = stroke_filename('/fn')
    log.set_stroke_filename(sf)
    log.enable_stroke_logging(True)
    log.stroke(Stroke(('S-', '-T', 'T-')))
    log.stroke(Stroke(('#', 'S-', '-T')))
    assert FakeHandler.outputs == {
        sf: ["Stroke(ST-T : ['S-', 'T-', '-T'])",
             "Stroke(1-9 : ['1-', '-9'])"],
    }

def test_log_translation():

    sf = stroke_filename('/fn')
    log.set_stroke_filename(sf)
    log.enable_translation_logging(True)
    log.translation(['a', 'b'], ['c', 'd'], None)
    assert FakeHandler.outputs == {sf: ['*a', '*b', 'c', 'd']}

def test_enable_stroke_logging():
    sf = stroke_filename('/fn')
    log.set_stroke_filename(sf)
    log.stroke(Stroke(('S-',)))
    log.enable_stroke_logging(True)
    log.stroke(Stroke(('T-',)))
    log.enable_stroke_logging(False)
    log.stroke(Stroke(('K-',)))
    assert FakeHandler.outputs == {sf: ["Stroke(T : ['T-'])"]}

def test_enable_translation_logging():
    sf = stroke_filename('/fn')
    log.set_stroke_filename(sf)
    log.translation(['a'], ['b'], None)
    log.enable_translation_logging(True)
    log.translation(['c'], ['d'], None)
    log.enable_translation_logging(False)
    log.translation(['e'], ['f'], None)
    assert FakeHandler.outputs == {sf: ['*c', 'd']}




import time

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import (
    QFileDialog,
    QFontDialog,
    QMessageBox,
)

from wcwidth import wcwidth

from plover import system

from plover.gui_qt.paper_tape_ui import Ui_PaperTape
from plover.gui_qt.i18n import get_gettext
from plover.gui_qt.utils import ToolBar
from plover.gui_qt.tool import Tool


_ = get_gettext()


class PaperTape(Tool, Ui_PaperTape):

    ''' Paper tape display of strokes. '''

    TITLE = _('Paper Tape')
    ICON = ':/tape.svg'
    ROLE = 'paper_tape'
    SHORTCUT = 'Ctrl+T'

    STYLE_PAPER, STYLE_RAW = (_('Paper'), _('Raw'))
    STYLES = (STYLE_PAPER, STYLE_RAW)

    def __init__(self, engine):
        super().__init__(engine)
        self.setupUi(self)
        self._strokes = []
        self._all_keys = None
        self._all_keys_filler = None
        self._formatter = None
        self._history_size = 2000000
        self.styles.addItems(self.STYLES)
        # Toolbar.
        self.layout().addWidget(ToolBar(
            self.action_ToggleOnTop,
            self.action_SelectFont,
            self.action_Clear,
            self.action_Save,
        ))
        self.action_Clear.setEnabled(False)
        self.action_Save.setEnabled(False)
        engine.signal_connect('config_changed', self.on_config_changed)
        self.on_config_changed(engine.config)
        engine.signal_connect('stroked', self.on_stroke)
        self.tape.setFocus()
        self.restore_state()
        self.finished.connect(self.save_state)

    def _restore_state(self, settings):
        style = settings.value('style', None, int)
        if style is not None:
            self.styles.setCurrentText(self.STYLES[style])
            self.on_style_changed(self.STYLES[style])
        font_string = settings.value('font')
        if font_string is not None:
            font = QFont()
            if font.fromString(font_string):
                self.header.setFont(font)
                self.tape.setFont(font)
        ontop = settings.value('ontop', None, bool)
        if ontop is not None:
            self.action_ToggleOnTop.setChecked(ontop)
            self.on_toggle_ontop(ontop)

    def _save_state(self, settings):
        settings.setValue('style', self.STYLES.index(self._style))
        settings.setValue('font', self.header.font().toString())
        ontop = bool(self.windowFlags() & Qt.WindowStaysOnTopHint)
        settings.setValue('ontop', ontop)

    def on_config_changed(self, config):
        if 'system_name' in config:
            self._strokes = []
            self._all_keys = ''.join(key.strip('-') for key in system.KEYS)
            self._all_keys_filler = [
                ' ' * wcwidth(k)
                for k in self._all_keys
            ]
            self._numbers = set(system.NUMBERS.values())
            self.header.setText(self._all_keys)
            self.on_style_changed(self._style)

    @property
    def _style(self):
        return self.styles.currentText()

    def _paper_format(self, stroke):
        text = self._all_keys_filler * 1
        keys = stroke.steno_keys[:]
        if any(key in self._numbers for key in keys):
            keys.append('#')
        for key in keys:
            index = system.KEY_ORDER[key]
            text[index] = self._all_keys[index]
        return ''.join(text)

    def _raw_format(self, stroke):
        return stroke.rtfcre

    def _show_stroke(self, stroke):
        text = self._formatter(stroke)
        self.tape.appendPlainText(text)

    def on_stroke(self, stroke):
        assert len(self._strokes) <= self._history_size
        if len(self._strokes) == self._history_size:
            self._strokes.pop(0)
        self._strokes.append(stroke)
        self._show_stroke(stroke)
        self.action_Clear.setEnabled(True)
        self.action_Save.setEnabled(True)

    def on_style_changed(self, style):
        assert style in self.STYLES
        if style == self.STYLE_PAPER:
            self.header.show()
            self._formatter = self._paper_format
        elif style == self.STYLE_RAW:
            self.header.hide()
            self._formatter = self._raw_format
        self.tape.clear()
        for stroke in self._strokes:
            self._show_stroke(stroke)

    def on_select_font(self):
        font, ok = QFontDialog.getFont(self.header.font(), self, '',
                                       QFontDialog.MonospacedFonts)
        if ok:
            self.header.setFont(font)
            self.tape.setFont(font)

    def on_toggle_ontop(self, ontop):
        flags = self.windowFlags()
        if ontop:
            flags |= Qt.WindowStaysOnTopHint
        else:
            flags &= ~Qt.WindowStaysOnTopHint
        self.setWindowFlags(flags)
        self.show()

    def on_clear(self):
        flags = self.windowFlags()
        msgbox = QMessageBox()
        msgbox.setText(_("Do you want to clear the paper tape?"))
        msgbox.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
        # Make sure the message box ends up above the paper tape!
        msgbox.setWindowFlags(msgbox.windowFlags() | (flags & Qt.WindowStaysOnTopHint))
        if QMessageBox.Yes != msgbox.exec_():
            return
        self._strokes = []
        self.action_Clear.setEnabled(False)
        self.action_Save.setEnabled(False)
        self.tape.clear()

    def on_save(self):
        filename_suggestion = 'steno-notes-%s.txt' % time.strftime('%Y-%m-%d-%H-%M')
        filename = QFileDialog.getSaveFileName(
            self, _('Save Paper Tape'), filename_suggestion,
            _('Text files') + ' (*.txt)',
        )[0]
        if not filename:
            return
        with open(filename, 'w') as fp:
            fp.write(self.tape.toPlainText())



#!/usr/bin/env python3

import os
import subprocess
import sys


# Default directory for caching wheels.
WHEELS_CACHE = os.path.join('.cache', 'wheels')


def _split_opts(text):
    args = text.split()
    assert (len(args) % 2) == 0
    return {name: int(nb_args)
            for name, nb_args
            in zip(*([iter(args)] * 2))}


# Allowed `pip install/wheel` options.
_PIP_OPTS = _split_opts(
    # General.
    '''
    --isolated 0
    -v 0 --verbose 0
    -q 0 --quiet 0
    --log 1
    --proxy 1
    --retries 1
    --timeout 1
    --exists-action 1
    --trusted-host 1
    --cert 1
    --client-cert 1
    --cache-dir 1
    --no-cache-dir 0
    --disable-pip-version-check 0
    --progress-bar 1
    '''
    # Install/Wheel.
    '''
    -c 1 --constraint 1
    -r 1 --requirement 1
    --no-deps 0
    '''
    # Package Index.
    '''
    -i 1 --index-url 1
    --extra-index-url 1
    --no-index 0
    -f 1 --find-links 1
    --process-dependency-links 0
    '''
)

# Allowed `pip install` only options.
_PIP_INSTALL_OPTS = _split_opts(
    '''
    -t 1 --target 1
    -U 0 --upgrade 0
    --upgrade-strategy 1
    --force-reinstall 0
    -I 0 --ignore-installed 0
    --user 0
    --root 1
    --prefix 1
    --require-hashes 0
    '''
)

_PIP_SCRIPT = '''
import sys

from pkg_resources import load_entry_point

# Work around isatty attribute being read-only with Python 2.
class NoTTY:
    def __init__(self, fp):
        self._fp = fp
    def isatty(self):
        return False
    def __getattr__(self, name):
        return getattr(self._fp, name)

if len(sys.argv) > 1 and sys.argv[1] == '--no-tty':
    sys.stdout = NoTTY(sys.stdout)
    del sys.argv[1]
sys.exit(load_entry_point('pip', 'console_scripts', 'pip')())
'''

def _pip(args, verbose=True, no_progress=True):
    if verbose:
        print('running pip %s' % ' '.join(a for a in args))
        sys.stdout.flush()
    cmd = [sys.executable, '-c', _PIP_SCRIPT]
    if no_progress:
        cmd.append('--no-tty')
    cmd.extend(args)
    return subprocess.call(cmd)

def install_wheels(args, verbose=True, no_install=False, no_progress=True):
    wheels_cache = WHEELS_CACHE
    wheel_args = []
    install_args = []
    constraint_args = []
    while len(args) > 0:
        a = args.pop(0)
        if not a.startswith('-'):
            wheel_args.append(a)
            install_args.append(a)
            continue
        if '=' in a:
            a = a.split('=', 1)
            args.insert(0, a[1])
            a = a[0]
        opt = a
        if opt == '-w':
            wheels_cache = args.pop(0)
            continue
        if opt == '--no-install':
            no_install = True
            continue
        if opt == '--progress-bar':
            if args[0] == 'off':
                no_progress = True
                del args[0]
                continue
            no_progress = False
        if opt in _PIP_OPTS:
            nb_args = _PIP_OPTS[opt]
            install_only = False
        elif opt in _PIP_INSTALL_OPTS:
            nb_args = _PIP_INSTALL_OPTS[opt]
            install_only = True
        else:
            raise ValueError('unsupported option: %s' % opt)
        a = [opt] + args[:nb_args]
        del args[:nb_args]
        if opt in ('-c', '--constraint'):
            constraint_args.extend(a)
            continue
        if not install_only:
            wheel_args.extend(a)
        install_args.extend(a)
    wheel_args[0:0] = ['wheel', '-f', wheels_cache, '-w', wheels_cache]
    install_args[0:0] = ['install', '-f', wheels_cache]
    if constraint_args:
        # Since pip will not build and cache wheels for VCS links, we do it ourselves:
        # - first, update the cache (without VCS links), and try to install from it
        code = _pip(wheel_args, no_progress=no_progress)
        if code == 0 and not no_install:
            code = _pip(install_args, no_progress=no_progress)
    else:
        code = 1
    if code != 0:
      # - if it failed, try to update the cache again, this time with VCS links
      code = _pip(wheel_args + constraint_args, no_progress=no_progress)
      if code == 0 and not no_install:
          # - and try again
          code = _pip(install_args, no_progress=no_progress)
    if code != 0:
        raise Exception('wheels installation failed: pip execution returned %u' % code)


if __name__ == '__main__':
    install_wheels(sys.argv[1:])



# Copyright (c) 2010-2011 Joshua Harlan Lifton.
# See LICENSE.txt for details.

"""Plover: Open Source Stenography Software"""

__version__ = '4.0.0.dev8'
__copyright__ = '(C) Open Steno Project'
__url__ = 'http://www.openstenoproject.org/'
__download_url__ = 'http://www.openstenoproject.org/plover'
__credits__ = ['Founded by stenographer Mirabai Knight.\n\nDevelopers:'
               '\nJoshua Lifton',
               'Hesky Fisher',
               'Ted Morin',
               'Benoit Pierre',
               '\nand many more on GitHub: \n<https://github.com/openstenoproject/plover>'
               ]
__license__ = 'GNU General Public License v2 or later (GPLv2+)'
__description__ = 'Open Source Stenography Software'
__long_description__ = """\
Plover is a free open source program intended to bring realtime
stenographic technology not just to stenographers, but also to
hackers, hobbyists, accessibility mavens, and all-around speed demons."""



# -*- coding: utf-8 -*-
# Copyright (c) 2010 Joshua Harlan Lifton.
# See LICENSE.txt for details.

"For use with a computer keyboard (preferably NKRO) as a steno machine."

from plover.machine.base import StenotypeBase
from plover.misc import boolean
from plover.oslayer.keyboardcontrol import KeyboardCapture


class Keyboard(StenotypeBase):
    """Standard stenotype interface for a computer keyboard.

    This class implements the three methods necessary for a standard
    stenotype interface: start_capture, stop_capture, and
    add_callback.

    """

    KEYS_LAYOUT = KeyboardCapture.SUPPORTED_KEYS_LAYOUT
    ACTIONS = ('arpeggiate',)

    def __init__(self, params):
        """Monitor the keyboard's events."""
        super().__init__()
        self._arpeggiate = params['arpeggiate']
        self._is_suppressed = False
        # Currently held keys.
        self._down_keys = set()
        # All keys part of the stroke.
        self._stroke_keys = set()
        self._keyboard_capture = None
        self._last_stroke_key_down_count = 0
        self._stroke_key_down_count = 0
        self._update_bindings()

    def _suppress(self):
        if self._keyboard_capture is None:
            return
        suppressed_keys = self._bindings.keys() if self._is_suppressed else ()
        self._keyboard_capture.suppress_keyboard(suppressed_keys)

    def _update_bindings(self):
        self._arpeggiate_key = None
        self._bindings = dict(self.keymap.get_bindings())
        for key, mapping in list(self._bindings.items()):
            if 'no-op' == mapping:
                self._bindings[key] = None
            elif 'arpeggiate' == mapping:
                if self._arpeggiate:
                    self._bindings[key] = None
                    self._arpeggiate_key = key
                else:
                    # Don't suppress arpeggiate key if it's not used.
                    del self._bindings[key]
        self._suppress()

    def set_keymap(self, keymap):
        super().set_keymap(keymap)
        self._update_bindings()

    def start_capture(self):
        """Begin listening for output from the stenotype machine."""
        self._initializing()
        try:
            self._keyboard_capture = KeyboardCapture()
            self._keyboard_capture.key_down = self._key_down
            self._keyboard_capture.key_up = self._key_up
            self._suppress()
            self._keyboard_capture.start()
        except:
            self._error()
            raise
        self._ready()

    def stop_capture(self):
        """Stop listening for output from the stenotype machine."""
        if self._keyboard_capture is not None:
            self._is_suppressed = False
            self._suppress()
            self._keyboard_capture.cancel()
            self._keyboard_capture = None
        self._stopped()

    def set_suppression(self, enabled):
        self._is_suppressed = enabled
        self._suppress()

    def suppress_last_stroke(self, send_backspaces):
        send_backspaces(self._last_stroke_key_down_count)
        self._last_stroke_key_down_count = 0

    def _key_down(self, key):
        """Called when a key is pressed."""
        assert key is not None
        self._stroke_key_down_count += 1
        self._down_keys.add(key)
        self._stroke_keys.add(key)

    def _key_up(self, key):
        """Called when a key is released."""
        assert key is not None
        self._down_keys.discard(key)
        # A stroke is complete if all pressed keys have been released,
        # and — when arpeggiate mode is enabled — the arpeggiate key
        # is part of it.
        if (
            self._down_keys or
            not self._stroke_keys or
            (self._arpeggiate and self._arpeggiate_key not in self._stroke_keys)
        ):
            return
        self._last_stroke_key_down_count = self._stroke_key_down_count
        steno_keys = {self._bindings.get(k) for k in self._stroke_keys}
        steno_keys -= {None}
        if steno_keys:
            self._notify(steno_keys)
        self._stroke_keys.clear()
        self._stroke_key_down_count = 0

    @classmethod
    def get_option_info(cls):
        return {
            'arpeggiate': (False, boolean),
        }



# -*- coding: utf-8 -*-
# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"""Unit tests for json.py."""

import inspect

import pytest

from plover.dictionary.json_dict import JsonDictionary

from .utils import make_dict
from . import parametrize


LOAD_TESTS = (
    lambda: ('{"S": "a"}', {('S', ): 'a'}),
    # Default encoding is utf-8.
    lambda: ('{"S": "café"}', {('S', ): 'café'}),
    # But if that fails, the implementation
    # must automatically retry with latin-1.
    lambda: ('{"S": "café"}'.encode('latin-1'), {('S', ): 'café'}),
    # Invalid JSON.
    lambda: ('{"foo", "bar",}', ValueError),
    # Invalid JSON.
    lambda: ('foo', ValueError),
    # Cannot convert to dict.
    lambda: ('"foo"', ValueError),
    # Ditto.
    lambda: ('4.2', TypeError),
)

@parametrize(LOAD_TESTS)
def test_load_dictionary(contents, expected):
    if isinstance(contents, str):
        contents = contents.encode('utf-8')
    with make_dict(contents) as filename:
        if inspect.isclass(expected):
            with pytest.raises(expected):
                JsonDictionary.load(filename)
        else:
            d = JsonDictionary.load(filename)
            assert dict(d.items()) == expected


SAVE_TESTS = (
    # Simple test.
    lambda: ({('S', ): 'a'},
             '{\n"S": "a"\n}'),
    # Check strokes format: '/' separated.
    lambda: ({('SAPL', '-PL'): 'sample'},
             '{\n"SAPL/-PL": "sample"\n}'),
    # Contents should be saved as UTF-8, no escaping.
    lambda: ({('S', ): 'café'},
             '{\n"S": "café"\n}'),
    # Check escaping of special characters.
    lambda: ({('S', ): '{^"\n\t"^}'},
             '{\n"S": "' + r'{^\"\n\t\"^}' + '"\n}'),
    # Keys are sorted on save.
    lambda: ({('B', ): 'bravo', ('A', ): 'alpha', ('C', ): 'charlie'},
             '{\n"A": "alpha",\n"B": "bravo",\n"C": "charlie"\n}'),
)

@parametrize(SAVE_TESTS)
def test_save_dictionary(contents, expected):
    with make_dict(b'foo') as filename:
        d = JsonDictionary.create(filename)
        d.update(contents)
        d.save()
        with open(filename, 'rb') as fp:
            contents = fp.read().decode('utf-8')
        assert contents == expected
